#!/usr/bin/env python

#
#
# Andrew Lukefahr
# lukefahr@indiana.edu
#
#

# Coerce Py2k to act more like Py3k
from __future__ import (absolute_import, division, print_function, unicode_literals)
from builtins import (
        ascii, bytes, chr, dict, filter, hex, input, int, isinstance, list, map,
        next, object, oct, open, pow, range, round, str, super, zip,
        )

import binascii
import logging
import os
import queue 
import socket
import struct
import sys
import threading
import time

from m3.m3_mbus import MBusInterface
from m3.m3_mbus import Memory
from m3.m3_mbus import RegFile

#inspired by
# https://github.com/0vercl0k/ollydbg2-python/blob/master/samples/gdbserver/gdbserver.py#L147

try: from . import m3_logging
except ValueError: import logging as m3_logging

class GdbRemote(object):
    '''
    This class handles the interface with the GDB client, and puts
    decoded messages in an output queue accessable through get()

    It also handles formatting responses to the client through the put() command

    It spawns a TX and RX thread internally
    '''

    class UnsupportedException(Exception): pass
    class PortTakenException(Exception): pass
    class DisconnectException(Exception): pass
    class CtrlCException(Exception): pass

    def __init__(this, tcp_port = 10001, log_level = logging.WARN):
        
        # setup our log
        this.log = m3_logging.getLogger( type(this).__name__)
        this.log.setLevel(log_level)
        
        #open our tcp/ip socket
        this.sock = socket.socket( socket.AF_INET, socket.SOCK_STREAM)

        #Bind socket to local host and port
        assert( isinstance(tcp_port,int) )
        try:
            this.sock.bind( ('localhost', tcp_port) )
        except socket.error as msg:
            this.log.error('Bind to port: ' + str(tcp_port) + \
                            ' failed. Error Code : ' + \
                            str(msg[0]) + ' Message ' + msg[1] )
            raise this.PortTakenException()

        this.log.info( 'Bound to port: ' + str(tcp_port))
        
        # inter-thread queues
        this.respQ = queue.Queue()
        this.reqQ = queue.Queue()
    
    def get(this,):
        '''
        Get the next incomming gdb client message 
        Blocking
        '''
        while True:
            try: return this.reqQ.get(True, 10)
            except queue.Empty: pass

    def put(this,msg):
        '''
        Put something on the outgoing (response) stream
        '''
        this.respQ.put(msg)

    def run(this,): 
        '''
        Starts the GDB server frontend
        (non-blocking call)
        '''
        this.RxTid = threading.Thread( target=this._gdb_rx, )
        this.RxTid.daemon = True
        this.RxTid.start()
        this.log.debug("Started GDB Thread")
        
    def _gdb_rx(this):
        ''' 
        Gdb client connection thread
        This thread handles accepting connections and incomming (receive) data
        It spawns a seperate thread for transmitting data
        '''
        this.sock.listen(1) #not sure why 1
        
        while True:
            [conn, client] = this.sock.accept()
            this.log.info('New connection from: ' + str(client[0]))
            
            #grab the opening '+'
            this.log.debug('Grabbing opening +')
            plus = conn.recv(1)
            assert(plus == '+')

            # start a response thread
            TxTid = threading.Thread( target=this._gdb_tx, args=(conn,) )
            TxTid.daemon = True
            TxTid.start()

            while True:
                try:
                    msg  = this._gdb_recv( conn)
                    if msg: 
                        this._process_command(msg)

                except this.CtrlCException:
                    this.log.debug('Caught CTRL+C')
                    this._gdbPut('_ctrlc_')

                except this.DisconnectException: 
                    # this is here to make testing easier, but isn't required
                    this._gdbPut('k')
                    this._gdbPut('_quit_')
                    this.put('GDB_QUIT')
                    break
            
            this.log.info('Closing connection with: ' + str(client[0]))
            conn.close()
            TxTid.join()

    def _gdb_tx(this, conn):
        ''' 
        Gdb Server Transmit thread
        runs until it pulls a 'GDB_QUIT' message from the incomming queue
        '''
        while True:
            msg = this._gdbGet()
            if msg == 'GDB_QUIT': 
                return
            elif msg == '+':
                this.log.debug('TX: ' + str(msg))
                conn.send(msg)
            else:
                this._gdb_resp(conn, msg) 

    def _gdbPut(this, cmd, *args, **kwargs):
        '''
        place a message on the to-be-processed queue
        '''
        this.reqQ.put( (cmd, args, kwargs) )
    
    def _gdbGet(this, timeout=None):
        '''
        get the next response from the queue
        (Blocking)
        '''
        ztime = 10 if timeout == None else timeout
        while True:
            try: return this.respQ.get(True, ztime)
            except queue.Empty: 
                if timeout == None: continue
                else: return None
    
 
    def _process_command(this,cmd):
        '''
        initial processing of a GDB client command
        mostly just puts it on the outgoing queue
        '''
        assert(len(cmd) > 0)

        cmdType = cmd[0]
        subCmd = cmd[1:]
       
        if cmdType == '+': return
        if cmdType == '?': 
            this._gdbPut('_question_')
        elif cmdType in [ 'D', 'c', 'k', 'g', 's' ]: 
            this._gdbPut(cmdType)
        elif cmdType in [ 'M', 'P', 'X', 'Z', 'm', 'p', 'q', 'v', 'z' ]: 
            this._gdbPut(cmdType, subCmd)
        elif cmdType in [ 'H', ]: 
            this._unsupported(cmdType, subCmd)
        else: raise this.UnsupportedException( cmdType)

    def _unsupported(this, cmdType, subCmd):
        '''
        Generic unsupported message response
        '''
        this.log.info("unsupported Type:" + str(cmdType))
        this.log.debug("SubCommand:" + str(subCmd))
        this.put("") #by-pass control thread

    
    def _gdb_recv(this, conn):
        '''
        receives and formats incomming messages
        @conn the connection to the gdb client 
        @throws DisconnectException the connection is disconnected
        @throws CtrlCException the message is 0x03 (CTRL-C)
        '''
    
        while True:
            rawdata= conn.recv(1024)
                    
            if not rawdata:
                raise this.DisconnectException()

            this.log.debug('RX: ' + str(rawdata) ) 

            # CTRL+C
            if chr(0x03) in rawdata:
                raise this.CtrlCException()
            
                # static buffer to tack on the new data
            # (plus fun way to make a static-ish function variable)
            try: this._buf_data += rawdata
            except AttributeError: this._buf_data = rawdata

            # acks "+" at the beginning can be safely removed
            if this._buf_data[0] == '+':
                this._buf_data= this._buf_data[1:]
        
            msg = None
            
            chkIdx = this._buf_data.find('#')
            # look for a checksum marker + 2 checksum bytes
            if (chkIdx > 0) and (len(this._buf_data) >= chkIdx + 3):
                #this.log.debug('Found # at: ' + str(chkIdx) )

                # get the message and checksum
                assert(this._buf_data[0] == '$')
                msg = this._buf_data[1:chkIdx]
                msgSum = int(this._buf_data[chkIdx+1:chkIdx+3],16)

                calcSum= 0
                for byte in msg:
                    calcSum = (calcSum + ord(byte)) & 0xff

                if calcSum != msgSum:
                    raise Exception("Checksum Error")
                else:
                    #this.log.debug('Checksum pass')
                    pass
                
                if '}' in msg:
                    raise Exception("FIXME: escape sequence")

                this._buf_data = this._buf_data[chkIdx+3:]

                this.log.debug('Parsed message : ' + str(msg) ) 
                #this.log.debug('Advanced buffered data: ' + \
                    # str(this._buf_data) ) 

                #ack message
                this.put('+') # bypass CTRL 

            return msg

    def _gdb_resp(this, conn, msg):
        '''
        takes a raw msg and formats it, and sends it over the connection
        to the GDB client 

        should not be used for '+' messages
        '''
        # calc checksum
        chkSum = 0
        for c in msg:
            chkSum += ord(c)
        chkSum = chkSum & 0xff
        
        if '}' in msg:
            raise Exception("FIXME: escape sequence")

        gdb_msg = '$%s#%.2x' % (msg, chkSum)
       
        this.log.debug('TX: ' + str(gdb_msg))
        conn.send( gdb_msg )



class GdbCtrl(object):
    '''
    The backend controller that actually impliments the GDB commands
    on the PRC over MBus
    '''

    class PrcCtrl(object):
        '''
        Manages the PRC, allows for out-of-band halt responses
        '''
        
        class HaltMBusInterface(MBusInterface):
            ''' 
            Overloads the MBus interface, allowing a special _callback
            '''
            def __init__(this, halt, _ice, prc_addr, log_level):
                super( halt.HaltMBusInterface, this).__init__(_ice, prc_addr,
                                                        log_level)
                this.halt = halt

            def _callback(this,*args, **kwargs):
                this.callback_queue.put((time.time(), args, kwargs))
                # only put special messages in the halt.queue...
                mbus_addr, mbus_data = args     
                [mbus_addr_int] = struct.unpack(">I", mbus_addr)
                if mbus_addr_int == 0xe0:
                    this.halt.queue.put( (mbus_addr, mbus_data) )
         
        class HaltRegFile(RegFile):
            ''' 
            Overloads the RegFile interface, 
            allowing thread-safeness for base_addr updates
            '''
            def __init__(this, halt, mbus, base_addr, writeback, log_level):
                super(halt.HaltRegFile, this).__init__(mbus, base_addr, \
                                        writeback, log_level)
                this.halt = halt
            def update_base_addr(this, base_addr):
                with this.halt.lock:
                    return RegFile.update_base_addr(this,base_addr)
            def __getitem__(this,key):
                assert(this.halt.flag_addr != None)
                with this.halt.lock:
                    return RegFile.__getitem__(this,key)
            def __setitem__(this,key,val):
                assert(this.halt.flag_addr != None)
                with this.halt.lock:
                    return RegFile.__setitem__(this,key,val)
            def forceWrite(this,key,val):
                assert(this.halt.flag_addr != None)
                with this.halt.lock:
                    return RegFile.forceWrite(this,key,val)

        def __init__(this, _ice, prc_addr, \
                                        log_level = logging.WARN):
            ''' 
            Manages the PRC through MBus.  

            Provides a memory and register file interface

            Also starts a halt-monitoring thread for out-of-band halt responses
            '''
            this.log = m3_logging.getLogger( type(this).__name__)
            this.log.setLevel(log_level)
            
            # the addr of the interrupt flag, None = not interrupted
            this.flag_addr = None

            # what we call when halt is triggered
            this.halt_cb = None

            this.lock = threading.RLock()
            this.queue = queue.Queue() # incomming PRC halt messages
            this.stop = threading.Event() # set this to stop the halt thread

            this.tid= threading.Thread( target=this._halt_thread, )
            this.tid.daemon = True
            this.tid.start()
            this.log.debug("Started Halt Thread")
            
            #our standard MBUS interfaces
            this.mbus = this.HaltMBusInterface( this, _ice, prc_addr, \
                                    log_level=log_level)
            this.mem = Memory(this.mbus, writeback=False, \
                                    log_level=log_level)
            this.rf = this.HaltRegFile(this,this.mbus,None,writeback=False, \
                                    log_level=log_level)
            this.log.debug("Created MBus/RF/Mem interfaces")
        
        def getMem(this): 
            ''' returns the PRC's memory abstraction'''
            return this.mem
        def getRF(this): 
            ''' returns the (thread-safe) PRC's register abstraction'''
            return this.rf
       
        def halt(this, halt_cb):
            '''
            non-blocking request to halt
            will call halt_cb("S05") when halt is complete
             possibly overwriting any existing halt_cb
            '''
            with this.lock:
                this.log.debug("Halting the PRC")
                assert(this.flag_addr == None)
                if (this.halt_cb):
                    this.log.debug("Overwriting existing halt_cb")
                this.halt_cb = halt_cb
                this.mbus.write_reg( 0x7, 0x1) # write something to MBUS_R7
                this.flag_addr = None
        
        def resume(this, halt_cb=None):
            '''
            resume the halted PRC
            will call halt_cb("S05") when halted if argument is given
            '''
            with this.lock:
                assert(this.flag_addr != None)
                assert(this.halt_cb == None)
                this.halt_cb = halt_cb
                this.log.debug("Resuming the PRC")
                this.log.debug("clearing flag @" + hex(this.flag_addr))
                this.mbus.write_mem(this.flag_addr, 0x01, 32)
                this.flag_addr = None
        
        def isHalted(this):
            ''' returns true is the PRC is currently soft-halted '''
            return (this.flag_addr != None)

        def _halt_thread(this):
            ''' 
            halt management thread, waits for halt to be triggered, 
            then calls halt_cb("S05") if halt_cb is valid 
            '''
            while not this.stop.isSet():
                try:  
                    mbus_addr, mbus_data = this.queue.get( True, 10)
                    this.log.debug("HALT triggered")

                    # read the gdb_flag and register pointer
                    [mbus_addr] = struct.unpack(">I", mbus_addr)
                    assert( mbus_addr == 0xe0)
                    [flag_addr ] = struct.unpack(">I", mbus_data)
                    this.log.debug("flag triggered")
                    this.log.debug("flag at: " + hex(flag_addr))
                except queue.Empty: continue

                try:
                    mbus_addr, mbus_data = this.queue.get()
                    [mbus_addr] = struct.unpack(">I", mbus_addr)
                    assert( mbus_addr == 0xe0)
                    [reg_addr ] = struct.unpack(">I", mbus_data)
                    this.log.debug("updating regFile at: " + hex(reg_addr))
                except Exception as e:
                    print (e)
                    raise
                
                with this.lock: 
                    this.rf.update_base_addr(reg_addr)
                    this.flag_addr = flag_addr
                    this.log.debug("Responding via GDB with SIGTRAP")
                    try:
                        this.halt_cb("S05")
                    except TypeError:
                        this.log.debug("skipping halt callback, not registered?")
                    this.halt_cb = None


    def __init__(this, ice, frontend, prc_addr, log_level = logging.WARN):

        # setup our log
        this.log = m3_logging.getLogger( type(this).__name__)
        this.log.setLevel(log_level)
       
        this.fe = frontend # the gdb frontend
        
        # the processor interface, and it's memory + reg file
        this.prc= this.PrcCtrl( ice, prc_addr, log_level)
        this.mem = this.prc.getMem()
        this.rf = this.prc.getRF()

        this.svc_01 = 0xdf01 # asm("SVC #01")

        # were displaced instructions live
        # these have the form { (addr,size) : inst }
        this.displaced_insts = {} 

        # try to import PyMulator (used to fake single-stepping)
        try:
            from PyMulator.PyMulator import PyMulator
            this.mulator = PyMulator(this.rf, this.mem,debug=True)
        except ImportError: 
            this.mulator = None
            this.log.warn('='*40 + '\n' + \
                         '\tPyMulator not found\n' +\
                         '\tSingle-stepping will not work!\n' + \
                         '\t Please install PyMulator:\n' +\
                         '\t $ pip install PyMulator\n' +\
                         '='*40)
                                         

        this.regs = [   'r0', 'r1', 'r2', 'r3', 'r4', 'r5', 'r6', 
                        'r7', 'r8', 'r9', 'r10', 'r11', 'r12', 'sp', 
                        'lr', 'pc', 
                        'f0', 'f1', 'f2', 'f3', 'f4', 'f5', 'f6', 'f7',                                 'fps', 
                        'xpsr', ]
        # how much 0-padding to put in front of a register 
        this.regsPads= { 'r0':0, 'r1':0, 'r2':0, 'r3':0, 'r4':0, 
                        'r5':0, 'r6':0, 'r7':0, 'r8':0, 'r9':0, 
                        'r10':0, 'r11':0, 'r12':0, 'sp':0, 'lr':0, 
                        'pc':0, 
                        'f0':8, 'f1':8, 'f2':8, 'f3':8, 'f4':8, 
                        'f5':8, 'f6':8, 'f7':8, 'fps':0, 
                        'xpsr':0, }

        this.encode_str = { 4:'<I', 2:'<H', 1:'<B' }
        
    def cmd__question_(this,):
        this.log.info("? Cmd")
        if this.prc.isHalted():
            return 'S05'
        else:
            this.cmd__ctrlc_()

    def cmd__ctrlc_(this,):
        this.log.info("CTRL+C (HALT)")
        this.prc.halt( this.fe.put )

    def cmd_D(this):
        this.log.info("detach")

        #halt the PRC if not already
        if not this.prc.isHalted(): 
            this.prc.halt(None)
        while not this.prc.isHalted():
            time.sleep(0.5)
        
        # remove all breakpoints, clear the table
        this.log.debug('removing breakpoints')
        for (addr,size) in this.displaced_insts:
            orig_inst = this.displaced_insts[(addr,size)]
            this.mem.forceWrite( (addr,size), orig_inst)
        this.displaced_insts = {}
        
        #then resume the PRC
        this.log.debug('resuming PRC')
        this.prc.resume()
        
        return 'OK'

    def cmd_M(this, subcmd):
        preamble, data = subcmd.split(':')
        addr,size_bytes= preamble.split(',')
        addr,size_bytes = map(lambda x: int(x, 16), [addr, size_bytes])
        this.log.info('mem write: ' + hex(addr) + ' of ' + str(size_bytes))
        data = binascii.unhexlify(data) 
        
        while size_bytes > 0:
            b = struct.unpack("B", data[0])[0]
            this.log.debug('Writing ' + hex(b) + ' to ' \
                + hex(addr))

            #this is not the most efficient, but it works...
            this.mem.forceWrite((addr,8), b)

            size_bytes -= 1
            addr += 1
            data = data[1:]
        return 'OK'

    def cmd_P(this, subcmd):
        assert(this.prc.isHalted()) 
        reg,val = subcmd.split('=')
        reg = int(reg, 16)
        reg = this.regs[ reg]
        # fix endiananess, conver to int
        val = int(binascii.hexlify( binascii.unhexlify(val)[::-1]),16)
        this.log.info('register write :' + str(reg) + ' = ' + hex(val) )
        this.rf.forceWrite(reg,val)
        return "OK"

    def cmd_X(this, subcmd):
        this.log.info("Binary Memory Write not supported.")
        return ""

    def cmd_Z(this, subcmd):
        args = subcmd.split(',')
        brType, addr, size = map(lambda x: int(x,16), args)
        assert(brType == 0)
        assert(size == 2)
        this.log.info("Breakpoint Set, Type:" + str(brType) + \
                                " Addr: " + hex(addr))
        this.log.debug("Replacing instruction @" + \
                        hex(addr) + " with trap" )
        size *= 8 # convert to bits
        if (addr,size) in this.displaced_insts: 
            this.log.debug( hex(addr) + '('+str(size)+')' + \
                'already a soft-breakpoint')
        else:
            this.displaced_insts[(addr,size)] = \
                    this.mem[(addr,size)]
            this.mem.forceWrite( (addr,16), this.svc_01)
        return 'OK'


    def cmd_c(this):
        assert(this.prc.isHalted()) 
        this.log.info("continue")
        this.prc.resume( this.fe.put )

    def cmd_g(this, ):
        assert(this.prc.isHalted()) 
        this.log.info('read all regs')

        resp = ''
        for ix in range(0, len(this.regs)):
            val = this.cmd_p( this.regs[ix] )
            resp += val
        return resp

    def cmd_k(this):
        this.log.info("kill")
        # just resume the PRC?  
        if this.prc.isHalted():
            this.prc.resume()

    def cmd_m(this, subcmd):
        assert(this.prc.isHalted()) 
        args = subcmd.split(',')
        addr,size_bytes = map(lambda x: int(x, 16), args)

        this.log.info('mem read: ' + hex(addr) + ' of ' + \
                            str(size_bytes))

        resp = '' 
        while size_bytes > 0:
            read_bytes = 4 if size_bytes >4 else size_bytes
            encode_str = this.encode_str[read_bytes]
            val = this.mem[(addr,read_bytes * 8)]
            this.log.debug('mem read: ' + hex(addr) + ' ' + hex(val))
            val = struct.pack(encode_str, val).encode('hex')#lit endian
            resp += val
            addr += read_bytes
            size_bytes -= read_bytes
        return resp

                
    def cmd_p(this, subcmd):
        assert(this.prc.isHalted()) 
        reg = subcmd
        this.log.info('reg read: ' + str(reg))
        encode = this.encode_str[4]

        val = this.rf[reg]
        if reg == 'pc': val -= 4
        val = struct.pack(encode ,val).encode('hex') #lit endian 
        val = '00' * this.regsPads[reg] + val # add some front-padding
        return val

    def cmd_q(this, subcmd):
        this.log.info('Query')
        if subcmd.startswith('C'):
            # asking about current thread, 
            # again, what threads...
            return ""
        elif subcmd.startswith('fThreadInfo'):
            # info on threads? what threads?
            return ""
        elif subcmd.startswith('L'):
            # legacy form of fThreadInfo
            return ""
        elif subcmd.startswith('Attached'):
            # did we attach to a process, or spawn a new one
            # processes?
            return ""
        elif subcmd.startswith('Offsets'):
            # did we translate the sections vith virtual memory?
            # virtual memory?
            return ""
        elif subcmd.startswith('Supported'):
            # startup command
            this.log.debug('qSupported')
            return "PacketSize=4096"
        elif subcmd.startswith('Symbol'):
            # gdb is offering us the symbol table
            return "OK"
        elif subcmd.startswith('TStatus'):
            #this has to do with tracing, we're not handling that yet
            return ""
        else: raise this.UnsupportedException( subcmd)
       
    def cmd_s(this, ):
        assert(this.prc.isHalted()) 
        this.log.info('single-step ')

        # might have to temporarially replace a trap
        displaced_trap = False
        pc = this.rf['pc'] - 4
        
        if (pc,16) in this.displaced_insts:
            this.log.debug("Trap @ " + hex(pc) + \
                    ", but we need the inst, fixing...")
            this.cmd_z('0,' + hex(pc)[2:] + ',2')
            displaced_trap = True

        # if the prc is halted, the reg file is valid 
        if True:
            this.log.debug("Soft-Stepping with Mulator")
            try:
                this.mulator.stepi()
            except AttributeError: 
                this.log.error("Stepping without PyMulator")
                this.log.error("CRASHING")
                raise

            break_addr = this.rf.getLocal('pc') - 4
            this.log.debug("Next PC: " + hex(break_addr) )
       
        # insert soft-trap at next instruction
        this.cmd_Z('0,' + hex(break_addr)[2:] + ',2')
        
        #step to the soft-trap
        this.prc.resume()
        while not this.prc.isHalted():
            time.sleep(0.1)
        
        #fix the next instruction
        this.cmd_z('0,' + hex(break_addr)[2:] + ',2')
        #and the orig inst
        if displaced_trap:
            this.cmd_Z('0,' + hex(pc)[2:] + ',2')

        return 'S05'

    def cmd_v(this, subcmd):
        this.log.info('v command: ' + str(subcmd))
        if subcmd.startswith('Cont?'):
            assert(this.prc.isHalted() )
            return "vCont;cs"
        elif subcmd.startswith('MustReplyEmpty'):
            return ""
        else: 
            this.log.warn("Unrecognized v command")
            return ""

    def cmd_z(this, subcmd):
        args = subcmd.split(',')
        brType, addr, size = map(lambda x: int(x,16), args)
        assert(brType == 0)
        assert(size == 2)
        this.log.info("Breakpoint Clear, Type: " + str(brType) + \
                        " Addr:" + hex(addr))
        this.log.debug("Replacing trap with origional instruction @" +\
                        hex(addr) )
        size *= 8 # convert to bites                                
        if (addr,size) not in this.displaced_insts:
            this.log.info( hex(addr) + '('+str(size)+')' + \
                'not a soft-breakpoint')
        else:
            orig_inst = this.displaced_insts[(addr,size)]
            this.mem.forceWrite( (addr,size), orig_inst)
            del this.displaced_insts[(addr,size)]
        return 'OK'



#
#
#
class test_GdbCtrl(GdbCtrl):
    ''' 
    A testing interface for the GDB Server frontend
    Just reports dummy values
    '''
    
    def __init__(this):

        this.log = m3_logging.getLogger( type(this).__name__)
        this.log.setLevel(logging.DEBUG)
        this.regs = [   'r0', 'r1', 'r2', 'r3', 'r4', 'r5', 'r6', 'r7', 'r8', 
                        'r9', 'r10', 'r11', 'r12', 'sp', 'lr', 'pc', 
                        'f0', 'f1', 'f2', 'f3', 'f4', 'f5', 'f6', 'f7', 'fps', 
                        'xpsr', ]
        this.regsPads= { 'r0':0, 'r1':0, 'r2':0, 'r3':0, 'r4':0, 'r5':0, 
                        'r6':0, 'r7':0, 'r8':0, 'r9':0, 'r10':0, 'r11':0, 
                        'r12':0, 'sp':0, 'lr':0, 'pc':0, 
                        'f0':8, 'f1':8, 'f2':8, 'f3':8, 'f4':8, 'f5':8, 'f6':8, 
                        'f7':8, 'fps':0, 
                        'xpsr':0, }

    def cmd__question_(this,):
        this.log.info("? Cmd")
        return 'S05'

    def cmd__ctrlc_(this,):
        this.log.info("CTRL+C (HALT)")
        #hack for now...
        return this.cmd__question_()

    def cmd_D(this):
        this.log.info("detach")
        return 'OK'

    def cmd_M(this, subcmd):
        preamble, data = subcmd.split(':')
        addr,size_bytes= preamble.split(',')
        addr,size_bytes = map(lambda x: int(x, 16), [addr, size_bytes])
        this.log.info('mem write: ' + hex(addr) + ' of ' + str(size_bytes))
        data = binascii.unhexlify(data) 
        while size_bytes > 0:
            b = struct.unpack("B", data[0])[0]
            this.log.debug('Writing ' + hex(b) + ' to ' \
                + hex(addr))

            size_bytes -= wr_size
            addr += wr_size 
            data = data[wr_size:]
        return 'OK'

    def cmd_P(this, subcmd):
        reg,val = subcmd.split('=')
        reg = int(reg, 16)
        reg = this.regs[ reg]
        # fix endiananess, conver to int
        val = int(binascii.hexlify( binascii.unhexlify(val)[::-1]),16)
        this.log.warn('register write :' + str(reg) + ' = ' + hex(val) )
        return "OK"

    def cmd_X(this, subcmd):
        this.log.info("Binary Memory Write not supported.")
        return ""

    def cmd_Z(this, subcmd):
        zType,addr,size= subcmd.split(',')
        this.log.info('breakpoint set: ' + addr + 'type: ' + zType )
        return 'OK'

    def cmd_c(this):
        this.log.info("continue")

    def cmd_g(this, ):
        this.log.debug('Read all Regs')
        resp = ''
        for ix in range(0, len(this.regs)):
            val = this.cmd_p( this.regs[ix] )
            resp += val
        return resp

    def cmd_k(this):
        this.log.info("kill")

    def cmd_m(this, subcmd):
        addr,size_bytes= subcmd.split(',')
        addr,size_bytes = map(lambda x: int(x, 16), [addr, size_bytes])
        this.log.info('mem read: ' + hex(addr) + ' of ' + str(size_bytes))
        if size_bytes == 4:
            return struct.pack('<I',0x46c046c0).encode('hex') #lit endian hex
        else: 
            return struct.pack('<H',0x46c).encode('hex') #lit endian hex

    def cmd_p(this, subcmd):
        reg = subcmd
        this.log.info('reg_read: ' + str(reg))
        val = 0x1234
        val = struct.pack('<I',val).encode('hex') #lit endian hex
        val = '00' * this.regsPads[reg] + val # add some front-padding
        return val

    def cmd_q(this, subcmd):
        this.log.debug('Query')
        if subcmd.startswith('C'):
            # asking about current thread, 
            # again, what threads...
            return ""
        elif subcmd.startswith('fThreadInfo'):
            # info on threads? what threads?
            return ""
        elif subcmd.startswith('L'):
            # legacy form of fThreadInfo
            return ""
        elif subcmd.startswith('Attached'):
            # did we attach to a process, or spawn a new one
            # processes?
            return ""
        elif subcmd.startswith('Offsets'):
            # did we translate the sections vith virtual memory?
            # virtual memory?
            return ""
        elif subcmd.startswith('Supported'):
            # startup command
            this.log.debug('qSupported')
            return "PacketSize=4096"
        elif subcmd.startswith('Symbol'):
            # gdb is offering us the symbol table
            return "OK"
        elif subcmd.startswith('TStatus'):
            #this has to do with tracing, we're not handling that yet
            return ""
        else: raise this.UnsupportedException( subcmd)


    def cmd_s(this, ):
        this.log.info('single-step ')
        return 'S05'
    
    def cmd_v(this, subcmd):
        if subcmd.startswith('Cont?'):
            this.log.debug('vCont')
            return "vCont;cs"
        else: assert(False) 

    def cmd_z(this, subcmd):
        ztype,addr,size= subcmd.split(',')
        this.log.info('breakpoint clear: ' + (addr))
        return 'OK'


if __name__ == '__main__':
   
    logging.basicConfig( level=logging.WARN, 
                            format='%(levelname)s %(name)s %(message)s')

    port = 10001

    if (len(sys.argv) > 1):
        for arg in sys.argv:
            if '--port=' in arg:
               port = int(arg.split("=")[1])
               print ('set port=' + str(port))

    ctrl = test_GdbCtrl()
    gdb = GdbRemote( tcp_port=port, log_level = logging.DEBUG)
    gdb.run()

    while True: 
        cmd, args, kwargs = gdb.get()
        cmd = 'cmd_'+cmd

        if cmd == 'cmd__quit_': 
            print ('GDB CTRL Quiting')
            break
        else : 
            func = getattr(ctrl, cmd)
            ret = func(*args, **kwargs)
            if ret != None: gdb.put(ret)




#!/usr/bin/env python

# Coerce Py2k to act more like Py3k
from __future__ import (absolute_import, division, print_function, unicode_literals)
from builtins import (
        ascii, bytes, chr, dict, filter, hex, input, int, isinstance, list, map,
        next, object, oct, open, pow, range, round, str, super, zip,
        )

import time
import os

from . import __version__

from .m3_common import m3_common
from .m3_common import mbus_snooper
from .m3_common import ein_programmer
from .m3_common import goc_programmer

from .m3_mbus import mbus_controller 

from . import m3_logging
logger = m3_logging.getLogger(__name__)

class m3_ice(m3_common):
    TITLE = "M3 ICE Interface"
    DESCRIPTION = "Tool to control the ICE board and attached M3 chips."
    MSG_TYPE = 'b+'

    def add_parse_args(self):
        super(m3_ice, self).add_parse_args()

        self.subparsers = self.parser.add_subparsers(
                title='Commands',
                description='Actions supported by the ICE board',
                )

        #aliases in Py3k only :(
        #self.parser_softreset = self.subparsers.add_parser('softreset',
        #        aliases=['reset'],
        #        help='Cycle 0.6V rail to reset M3 chips')
        self.parser_softreset = self.subparsers.add_parser('reset',
                parents=[self.parent_parser],
                help='Cycle 0.6V rail to reset M3 chips')
        self.parser_softreset.set_defaults(func=self.cmd_softreset)

        self.parser_hardreset = self.subparsers.add_parser('hardreset',
                parents=[self.parent_parser],
                help='Cycle all power rails to cold-boot M3 chips')
        self.parser_hardreset.set_defaults(func=self.cmd_hardreset)

        self.parser_power = self.subparsers.add_parser('power',
                parents=[self.parent_parser],
                help='Control power rails sent to connected M3 chips')
        self.parser_power.add_argument('STATE',
                choices=['on', 'off'],
                help='Power M3 chips on or off')
        self.parser_power.set_defaults(func=self.cmd_power)

        self.parser_snoop = self.subparsers.add_parser('snoop',
                parents=[self.parent_parser],
                help='Passively monitor MBus messages')
        mbus_snooper.add_parse_args(self.parser_snoop)
        self.parser_snoop.set_defaults(func=self.cmd_snoop)

        self.parser_ein = self.subparsers.add_parser('ein',
                parents=[self.parent_parser],
                help='Command the chip via the EIN protocol')
        ein_programmer.add_parse_args(self.parser_ein)
        self.parser_ein.set_defaults(func=self.cmd_ein)

        self.parser_goc = self.subparsers.add_parser('goc',
                parents=[self.parent_parser],
                help='Send commands via the GOC protocol (blinking light)')
        self.goc_programmer = goc_programmer(self, self.parser_goc)
        self.parser_goc.set_defaults(func=self.cmd_goc)

        self.parser_mbus = self.subparsers.add_parser('mbus',
                parents=[self.parent_parser],
                help='Send commands via the MBUS protocol')

        self.mbus_controller = mbus_controller(self,self.parser_mbus)


    def cmd_softreset(self):
        self.ice.power_set_onoff(self.ice.POWER_0P6, False)
        time.sleep(.5)
        self.ice.power_set_onoff(self.ice.POWER_0P6, True)
        logger.info("Soft reset complete.")

    def cmd_hardreset(self):
        self._cmd_power_off()
        time.sleep(2)

        self._cmd_power_on()
        time.sleep(4)

        self.cmd_softreset()

    def _cmd_power_off(self):
        self.ice.power_set_onoff(self.ice.POWER_0P6, False)
        self.ice.power_set_onoff(self.ice.POWER_1P2, False)
        self.ice.power_set_onoff(self.ice.POWER_VBATT, False)
        logger.info("Power off.")

    def _cmd_power_on(self):
        self.ice.power_set_onoff(self.ice.POWER_VBATT, True)
        self.ice.power_set_onoff(self.ice.POWER_1P2, True)
        self.ice.power_set_onoff(self.ice.POWER_0P6, True)
        logger.info("Power on.")

    def cmd_power(self):
        getattr(self, '_cmd_power_{}'.format(self.args.STATE))()

    def cmd_snoop(self):
        snooper = mbus_snooper(self.args, self.ice)
        self.hang_for_messages()

    def cmd_ein(self):
        ein = ein_programmer(self)
        ein.cmd()

    def cmd_goc(self):
        #goc_programmer(self).cmd()
        pass

    # now handled by mbus_controller internally 
    #def cmd_mbus(self):
    #    pass


def cmd():
    try:
        os.environ['ICE_DEBUG']
        print('M3 Library Version:', __version__)
    except KeyError:
        pass
    m = m3_ice()
    m.args.func()

if __name__ == '__main__':
    cmd()




#!/usr/bin/env python2
# vim: sts=4 ts=4 sw=4 noet:

import os
import logging

#from pdb import set_trace as bp

logging.captureWarnings(True)

TRACE_LEVEL=25
logging.addLevelName(TRACE_LEVEL, 'TRACE')

def split_line_logger(lvl, self, message, *args, **kwargs):
	#print('lvl: {}, message: {}'.format(lvl, message))
	for msg in message.split('\n'):
		if self.isEnabledFor(lvl):
			self._log(lvl, msg, args, **kwargs)

#for lvl,logger in (
#		(logging.DEBUG,   'debug'),
#		(logging.INFO,    'info'),
#		(logging.WARN,    'warn'),
#		(logging.ERROR,   'error'),
#		(logging.CRITICAL,'critical'),
#		(TRACE_LEVEL,     'trace'),
#		):
#	setattr(logging.Logger, logger, lambda self, message, *args, **kwargs :\
#			split_line_logger(lvl, self, message, *args, **kwargs))

setattr(logging.Logger, 'debug', lambda self, message, *args, **kwargs :\
		split_line_logger(logging.DEBUG, self, message, *args, **kwargs))
setattr(logging.Logger, 'info', lambda self, message, *args, **kwargs :\
		split_line_logger(logging.INFO, self, message, *args, **kwargs))
setattr(logging.Logger, 'warn', lambda self, message, *args, **kwargs :\
		split_line_logger(logging.WARN, self, message, *args, **kwargs))
setattr(logging.Logger, 'error', lambda self, message, *args, **kwargs :\
		split_line_logger(logging.ERROR, self, message, *args, **kwargs))
setattr(logging.Logger, 'critical', lambda self, message, *args, **kwargs :\
		split_line_logger(logging.CRITICAL, self, message, *args, **kwargs))
setattr(logging.Logger, 'trace', lambda self, message, *args, **kwargs :\
		split_line_logger(TRACE_LEVEL, self, message, *args, **kwargs))

class DefaultFormatter(logging.Formatter):
	def __init__(self, fmt="%(levelname)s\t%(message)s"):
		super(DefaultFormatter, self).__init__(fmt)

class NoTraceFilter(logging.Filter):
	def filter(self, record):
		return not record.levelno == TRACE_LEVEL

def is_lambda(fn):
	# This is a little fragile, but really this whole introspection idea is, so
	# let's roll with it for now
	return isinstance(fn, type(lambda: None)) and fn.__name__ == '<lambda>'

def fn_to_source(fn):
	if is_lambda(fn):
		source = inspect.getsource(fn).strip()
		source = source.replace('\\\n', '')
		source = source.replace('\n', '')
		source = source.replace('\t', ' ')
		while source.find('  ') != -1:
			source = source.replace('  ', ' ')
		if source.find('command') != -1:
			source = source[source.find('command'):]
		return source
	else:
		return fn.__name__

import inspect
def trace(logger, fn):
	"""Decorator that logger.trace()'s function calls and their arguments"""
	def inner(*args, **kwargs):
		source = fn_to_source(fn)
		logger.trace('{} *args={} **kwargs={}'.format(source, args, kwargs))
		#for frame in inspect.stack():
		#	logger.trace(frame)
		#	logger.trace(inspect.getframeinfo(frame[0]))
		#	logger.trace(inspect.getargvalues(frame[0]))
		#	logger.trace(' ')
		#logger.trace(inspect.getargvalues(inspect.trace()[0][0]))
		return fn(*args, **kwargs)
	return inner

no_trace_filter = NoTraceFilter()

class LoggerManager (object):
    
    def __init__(this):

        this.level = logging.DEBUG 

        this.f = DefaultFormatter()

        this.h = logging.StreamHandler()
        this.h.setFormatter(this.f)
        this.h.level = logging.INFO
        this.h.addFilter(NoTraceFilter())

        this.debugFormat = "%(asctime)s\t%(levelname)s\t%(name)s\t%(message)s"

    def getLogger(this, name):
        l = logging.getLogger(name)

        l.level = this.level
        l.addHandler(this.h)
        l.trace = lambda msg, l=l: l.log(TRACE_LEVEL, msg)

        return l

    def updateLevel(this, log_level):
       
        if isinstance(log_level, int): 
            this.h.level = log_level
            if log_level == logging.DEBUG:
                this.updateFormat(this.debugFormat)
            else: this.updateFormat()

        else: # if called with strings, recurse with the level
            if log_level.lower() in ['debug']:
                this.updateLevel(logging.DEBUG)
            elif log_level.lower() in ['info']:
                this.updateLevel(logging.INFO)
            elif log_level.lower() in ['warn']:
                this.updateLevel(logging.WARN)
            else: raise Exception("Bad Level")

    def updateFormat(this, fmt=None):
        if fmt!= None:
            this.f = DefaultFormatter(fmt=fmt)
        else:
            this.f = DefaultFormatter()
        this.h.setFormatter(this.f)

def log_level_from_environment():
    try:
        os.environ['ICE_DEBUG']
        return logging.DEBUG
    except KeyError:
        return logging.INFO

# logger manager
logMan = LoggerManager() 

#and a global logger
glob_logger = logMan.getLogger(__name__)

def getGlobalLogger():
    return glob_logger

def LoggerSetLevel(lvl):
    logMan.updateLevel(lvl)

def get_logger(name):
    return logMan.getLogger(name)

#maintain API compatability with the logger module
def getLogger(name):
    return logMan.getLogger(name)





#!/usr/bin/env python

# Coerce Py2k to act more like Py3k
from __future__ import (absolute_import, division, print_function, unicode_literals)
from builtins import (
        ascii, bytes, chr, dict, filter, hex, input, int, isinstance, list, map,
        next, object, oct, open, pow, range, round, str, super, zip,
        )

CAPABILITES = "?_dIifnOoBbMmeGgPp"
MAX_GPIO = 24
DEFAULT_BAUD_DIVIDER = 0x00AE
DEFAULT_I2C_MASK = '1001100x'
DEFAULT_I2C_SPEED_IN_KHZ = 100
DEFAULT_FLOW_CLOCK_IN_HZ = .625
DEFAULT_POWER_0P6 = 0.675
DEFAULT_POWER_1P2 = 1.2
DEFAULT_POWER_VBATT = 3.8
DEFAULT_VSET_0P6 = 19
DEFAULT_VSET_1P2 = 25
DEFAULT_VSET_VBATT = 25
DEFAULT_MBUS_FULL_PREFIX_ONES = 0xfffff0
DEFAULT_MBUS_FULL_PREFIX_ZEROS = 0xfffff0
DEFAULT_MBUS_SHORT_PREFIX = 0x0f
DEFAULT_MBUS_BROADCAST_MASK_ONES = 0x0f
DEFAULT_MBUS_BROADCAST_MASK_ZEROS = 0x0f
DEFAULT_MBUS_SNOOP_BROADCAST_MASK_ONES = 0x0f
DEFAULT_MBUS_SNOOP_BROADCAST_MASK_ZEROS = 0x0f


import argparse
import atexit
import binascii
import datetime
import os
import platform
import random
import serial
import struct
import subprocess
import sys
import tempfile
import time
import threading
import traceback


from . import m3_logging
logger = m3_logging.get_logger(__name__)

from .ice import ICE

class UnknownCommandException(Exception):
    pass


class Simulator(object):
    def __init__(self, args=None):
        if args is None:
            self.parse_cli()
        else:
            self.args = args

        self.baud_divider = DEFAULT_BAUD_DIVIDER

        self.i2c_mask_ones = 0
        self.i2c_mask_zeros = 0
        for bit in self.args.i2c_mask:
            self.i2c_mask_ones <<= 1
            self.i2c_mask_zeros <<= 1
            self.i2c_mask_ones |= (bit == '1')
            self.i2c_mask_zeros |= (bit == '0')
        logger.debug("mask %s ones %02x zeros %02x", self.args.i2c_mask,
                self.i2c_mask_ones, self.i2c_mask_zeros)

        self.i2c_speed_in_khz = DEFAULT_I2C_SPEED_IN_KHZ

        self.flow_clock_in_hz = DEFAULT_FLOW_CLOCK_IN_HZ
        self.flow_onoff = False

        self.vset_0p6 = DEFAULT_VSET_0P6
        self.vset_1p2 = DEFAULT_VSET_1P2
        self.vset_vbatt = DEFAULT_VSET_VBATT
        self.power_0p6_on = False
        self.power_1p2_on = False
        self.power_vbatt_on = False
        self.power_goc_on = False

        self.mbus_full_prefix_ones = DEFAULT_MBUS_FULL_PREFIX_ONES
        self.mbus_full_prefix_zeros = DEFAULT_MBUS_FULL_PREFIX_ZEROS
        self.mbus_short_prefix = DEFAULT_MBUS_SHORT_PREFIX
        self.mbus_snoop_enabled = False
        self.mbus_broadcast_mask_ones = DEFAULT_MBUS_BROADCAST_MASK_ONES
        self.mbus_broadcast_mask_zeros = DEFAULT_MBUS_BROADCAST_MASK_ZEROS
        self.mbus_snoop_broadcast_mask_ones = DEFAULT_MBUS_SNOOP_BROADCAST_MASK_ONES
        self.mbus_snoop_broadcast_mask_zeros = DEFAULT_MBUS_SNOOP_BROADCAST_MASK_ZEROS
        self.mbus_ismaster = False
        self.mbus_should_interrupt = 0
        self.mbus_should_prio = 0
        self.mbus_force_reset = 0


        self.s_lock = threading.Lock()
        self.s_en_event = threading.Event()

        logger.debug('Serial:' + str(self.args.serial))
        if self.args.serial == _FAKE_SERIAL_SIMULATOR_ENDPOINT:
            if not os.path.exists(_FAKE_SERIAL_SIMULATOR_ENDPOINT):
                create_fake_serial()

        try:
            if self.args.serial == _FAKE_SERIAL_SIMULATOR_ENDPOINT:
                # Workaround for https://github.com/pyserial/pyserial/issues/59
                #            and https://github.com/pyserial/pyserial/issues/113
                serial.Serial._update_dtr_state = lambda self : None
                serial.Serial._update_rts_state = lambda self : None
            self.s = serial.Serial(self.args.serial, 115200)
        except:
            logger.error("Opening serial failed")
            logger.error("")
            logger.error("If you need to create a software serial tunnel, use socat:")
            logger.error("  socat -x pty,link=/tmp/com1,raw,echo=0 pty,link=/tmp/com2,raw,echo=0")
            logger.error("")
            raise
        if not self.s.isOpen():
            logger.error('Could not open serial port at: ' + self.args.serial)
            raise IOError("Failed to open serial port")


        self.event = 0
        self.gpios = [Gpio() for x in range(MAX_GPIO)]

        if self.args.generate_messages:
            self.gen_thread = threading.Thread(target=self.spurious_message_thread)
            self.gen_thread.daemon = True
            self.gen_thread.start()

        if self.args.replay is not None:
            self.replay_thread = threading.Thread(target=self.replay_message_thread)
            self.replay_thread.daemon = True
            self.replay_thread.start()


    def run(self, background=False):
        if background:
            self.main_thread = threading.Thread(target=self.main_loop)
            self.main_thread.daemon = True
            self.main_thread.start()
        else:
            logger.info("-" * 80)
            logger.info("-- M3 ICE Interface Board Simulator")
            logger.info("")

            if self.args.transaction:
                #skip multithreading here
                self.transaction_mode()
            else:
                self.main_loop()


    def spurious_message_thread(self):
        def send_snoop(addr, data, control):
            logger.info("Send generated message ADDR: 0x{}   DATA: 0x{}   CTL: 0x{}".\
                    format(addr, data, control))
            with self.s_lock:
                self.s.write(bytes('B', 'utf-8'))
                self.s.write(bytes((self.event,)))
                self.event += 1
                self.event %= 256

                addr = binascii.unhexlify(addr)
                data = binascii.unhexlify(data)
                control = binascii.unhexlify(control)
                length = len(addr) + len(data) + len(control)

                self.s.write(bytes((length,)))
                self.s.write(addr)
                self.s.write(data)
                self.s.write(control)

        while True:
            # control:
            #   b0 b1 -> val -> meaning
            #  - 0, 0     0     General Error
            #  - 0, 1     2     TX or RX Error
            #  - 1, 0     1     ACK
            #  - 1, 1     3     NAK
            for args in (
                    ('00000074', 'deadbeef', '01'),
                    ('00000040', 'ab', '01'),
                    ('f0012345', '0123456789abcdef', '03'),
                    ('00000022', 'a5'*160, '01'),
                    ('00000033', 'c9'*160, '01'),
                    ('00000044', 'ef'*160, '01'),
                    ):
                self.sleep(random.randint(1,12))
                if not self.mbus_snoop_enabled:
                    continue
                send_snoop(*args)

    def replay_message_thread(self):
        def send_snoop(addr, data, control):
            with self.s_lock:
                self.s.write(bytes('B', 'utf-8'))
                self.s.write(bytes((self.event,)))
                self.event += 1
                self.event %= 256

                addr = binascii.unhexlify(addr)
                data = binascii.unhexlify(data)
                control = binascii.unhexlify(control)
                length = len(addr) + len(data) + len(control)

                self.s.write(bytes((length,)))
                self.s.write(addr)
                self.s.write(data)
                self.s.write(control)

        logger.info("Replay thread waiting for snoop to be enabled")
        self.s_en_event.wait()
        logger.info("Replay beginning")
        last_ts = None
        for line in open(self.args.replay):
            assert self.mbus_snoop_enabled

            ts,addr,data = line.strip().split(',')
            if len(addr) == 2:
                addr = '000000' + addr
            else:
                assert len(addr) == 8
            ts = float(ts)

            #if last_ts is not None:
            #    logger.info("sleep for {}".format(ts - last_ts))
            #    self.sleep(ts - last_ts)
            last_ts = ts

            print(data)
            print(len(data)/2.0)
            send_snoop(addr, data, '02')

        logger.info("Replay finished.")

    #
    #
    #
    def transaction_mode(self):
        ''' 
        Replays a series of ICE transactions with timing information
        '''
        def read_raw_message():
            msg_type, event_id, length = self.s.read(3)
            length_int = ord(length)
            logger.debug("Got a message of type: " + msg_type + 
                    ' length: ' + str(length_int))
            msg = self.s.read(length_int)

            return msg_type + event_id + length + msg

        logger.info("Transaction beginning")
        last_ts = None

        for line in open( self.args.transaction.strip() ):
            
            line = line.strip('\n')
            line = line.strip('\r')

            logger.info('Working on: ' + line)

            if len(line) == 0: continue
            elif line[0] in ['#', ' ', '/', ]: continue

            elif line.startswith('WAIT'):
               
                #find the number after T
                hex_tex = line.split('WAIT')[1].strip()
                hex_tex = hex_tex.replace('0x', '').lower()
                data = binascii.unhexlify(hex_tex)
               
                rxMsg = b''

                rxMsg = read_raw_message()
                logger.debug('Read: ' + binascii.hexlify(rxMsg))
                logger.info(' vs  : ' + binascii.hexlify(data))
                if (rxMsg != data): 
                    rx = binascii.hexlify(rxMsg)
                    buf = binascii.hexlify(data) 
                    raise Exception('Read vs. Expect: ' + \
                            str(rx) + ' vs. ' + str(buf)  + \
                            ' ascii: ' + str(rx==buf))
                logger.debug('Found match!') 

            elif line.startswith('SEND'):
                # find the number after N (D could also be hex...)
                hex_tex = line.split('SEND')[1].strip()
                hex_tex = hex_tex.replace('0x', '').lower()
                data = binascii.unhexlify(hex_tex)
                print ('SENDING: ' + binascii.hexlify(data))
                self.s.write(data)
                self.s.flush()

            elif line.startswith('SLEEP'): 
                time_str = line.replace(' ','').split('P')[1]
                time_sec = float(time_str)
                time.sleep(time_sec)
                self.s.flush()



            else: raise Exception('Unknown Command: "' + repr(line) + '"')

        time.sleep(0.5)
        self.s.close()

        logger.info("Transaction finished.")


    def main_loop(self):
        self.i2c_msg = bytes()
        self.i2c_match = True
        self.flow_msg = bytes()
        self.ein_msg = bytes()
        self.mbus_msg = bytes()
        while True:
            def min_proto(proto):
                if minor < proto:
                    logger.error("Request for protocol 0.2 command, but the")
                    logger.error("negotiated protocol was 0.1")
                    raise UnknownCommandException

            try:
                msg_type, event_id, length = self.s.read(3)
                logger.debug("Got a message of type: " + msg_type)
                event_id = ord(event_id)
                length = ord(length)
                msg = self.s.read(length)
    
                #slight hack to simplify respond()
                self.event = event_id

                if msg_type == 'V':
                    if self.args.ice_version == 1:
                        self.respond(binascii.unhexlify('0001'))
                    elif self.args.ice_version == 2:
                        self.respond(binascii.unhexlify('0001'))
                    elif self.args.ice_version == 3:
                        self.respond(binascii.unhexlify('000300020001'))
                    elif self.args.ice_version == 4:
                        self.respond(binascii.unhexlify('0004000300020001'))
                    else:
                        raise ValueError("Unknown ice version: %d" % (self.args.ice_version))
                elif msg_type == 'v':
                    CLOCK_FREQ = 4e6
                    if msg == binascii.unhexlify('0004'):
                        minor = 4
                    elif msg == binascii.unhexlify('0003'):
                        minor = 3
                    elif msg == binascii.unhexlify('0002'):
                        minor = 2
                    elif msg == binascii.unhexlify('0001'):
                        CLOCK_FREQ = 2e6
                        minor = 1
                    else:
                        logger.error("Request for unknown version: " + msg)
                        raise Exception
                    logger.info("Negotiated to protocol version 0."+ str(minor))
                    self.ack()

                elif msg_type == '?':
                    min_proto(2)
                    if msg[0] == '?':
                        logger.info("Responded to query capabilites with " + CAPABILITES)
                        self.respond(CAPABILITES)
                    elif msg[0] == 'b':
                        logger.info("Responded to query for ICE baudrate (divider: 0x%04X)" % (self.baud_divider))
                        self.respond(struct.pack('>H', self.baud_divider))
                    else:
                        logger.error("Bad '?' subtype: " + msg[0])
                        raise UnknownCommandException
                elif msg_type == '_':
                    min_proto(2)
                    if msg[0] == 'b':
                        high = ord(msg[1])
                        low = ord(msg[2])
                        new_div = low | (high << 8)
                        if new_div not in (0x00AE, 0x000A, 0x0007):
                            logger.error("Bad baudrate divider: 0x%04X" % (new_div))
                            raise Exception
                        self.ack()
                        try:
                            if new_div == 0x00AE:
                                self.s.baudrate = 115200
                            elif new_div == 0x000A:
                                self.s.baudrate = 2000000
                            elif new_div == 0x0007:
                                self.s.baudrate = 3000000
                            else:
                                logger.error("Unknown baudrate divider")
                                raise Exception
                        except IOError as e:
                            if e.errno == 25:
                                logger.warn("Failed to set baud rate (if socat, ignore)")
                            else:
                                raise
                        self.baud_divider = new_div
                        logger.info("New baud divider set: " + str(self.baud_divider))
                    else:
                        logger.error("bad '_' subtype: " + msg[0])
                        raise UnknownCommandException
                elif msg_type == 'b':
                    min_proto(2)
                    self.mbus_msg += msg
                    if len(msg) != 255:
                        logger.info("Got a MBus message:")
                        logger.info("   message: " + binascii.hexlify(self.mbus_msg))
                        self.mbus_msg = bytes()
                        if self.mbus_should_interrupt:
                            logger.info("Message would have interrupted")
                            if self.mbus_should_interrupt == 1:
                                self.mbus_should_interrupt = 0
                        if self.mbus_should_prio:
                            logger.info("Message would have been sent high priority")
                            if self.mbus_should_prio == 1:
                                self.mbus_should_prio = 0
                    else:
                        logger.debug("Got MBus fragment")
                    self.ack()
                elif msg_type == 'd':
                    self.i2c_msg += msg
                    if not self.i2c_match:
                        if not self.match_mask(ord(msg[0]), self.i2c_mask_ones, self.i2c_mask_zeros):
                            logger.info("I2C address %02x did not match mask %02x %02x",
                                    ord(msg[0]), self.i2c_mask_ones, self.i2c_mask_zeros)
                            self.respond(struct.pack('B', 0), ack=False)
                            continue
                        self.i2c_match = True
                    if len(msg) != 255:
                        logger.info("Got i2c message:")
                        logger.info("  addr: " + binascii.hexlify(self.i2c_msg[0:1]))
                        logger.info("  data: " + binascii.hexlify(self.i2c_msg[1:]))
                        self.i2c_msg = bytes()
                        self.i2c_match = False
                    else:
                        logger.debug("Got i2c fragment")
                    self.ack()
                elif msg_type == 'e':
                    min_proto(2)
                    self.ein_msg += msg
                    if len(msg) != 255:
                        logger.info("Got a EIN message:")
                        logger.info("  message: " + binascii.hexlify(self.ein_msg))
                        self.ein_msg = bytes()
                    else:
                        logger.debug("Got EIN fragment")
                    self.ack()
                elif msg_type in ('f', 'n'):
                    self.flow_msg += msg
                    if len(msg) != 255:
                        logger.info("Got f/n-type message in %s mode:", ('EIN','GOC')[ein_goc_toggle])
                        logger.info("  message: " + binascii.hexlify(self.flow_msg))
                        self.flow_msg = bytes()
                    else:
                        logger.debug("Got f/n-type fragment in %s mode", ('EIN','GOC')[ein_goc_toggle])
                    if ein_goc_toggle:
                        t = (len(msg)*8) / self.flow_clock_in_hz
                        logger.info("Sleeping for {} seconds to mimic GOC".format(t))
                        try:
                            self.sleep(t)
                        except KeyboardInterrupt:
                            pass
                    self.ack()
                elif msg_type == 'G':
                    # GPIO changed completely between v0.1 and v0.2
                    if minor == 1:
                        if msg[0] == 'l':
                            logger.info("Responded to request for GPIO %d Dir (%s)", ord(msg[1]), self.gpios[ord(msg[1])])
                            self.respond(struct.pack("B", self.gpios[ord(msg[1])].level))
                        elif msg[0] == 'd':
                            logger.info("Responded to request for GPIO %d Level (%s)", ord(msg[1]), self.gpios[ord(msg[1])])
                            self.respond(struct.pack("B", self.gpios[ord(msg[1])].direction))
                        else:
                            logger.error("bad 'G' subtype: " + msg[0])
                            raise Exception
                    else:
                        if msg[0] == 'l':
                            mask = 0
                            for i in range(len(self.gpios)):
                                mask |= (self.gpios[i].level << i)
                            logger.info("Responded to request for GPIO level mask (%06x)", mask)
                            self.respond(struct.pack('>I', mask)[1:])
                        elif msg[0] == 'd':
                            mask = 0
                            for i in range(len(self.gpios)):
                                mask |= (self.gpios[i].direction << i)
                            logger.info("Responded to request for GPIO direction mask (%06x)", mask)
                            self.respond(struct.pack('>I', mask)[1:])
                        elif msg[0] == 'i':
                            mask = 0
                            for i in range(len(self.gpios)):
                                mask |= (self.gpios[i].interrupt << i)
                            logger.info("Responded to request for GPIO interrupt mask (%06x)", mask)
                            self.respond(struct.pack('>I', mask)[1:])
                        else:
                            logger.error("bad 'G' subtype: " + msg[0])
                            raise Exception
                elif msg_type == 'g':
                    # GPIO changed completely between v0.1 and v0.2
                    if minor == 1:
                        if msg[0] == 'l':
                            self.gpios[ord(msg[1])].level = (ord(msg[2]) == True)
                            logger.info("Set GPIO %d Level: %s", ord(msg[1]), self.gpios[ord(msg[1])])
                            self.ack()
                        elif msg[0] == 'd':
                            self.gpios[ord(msg[1])].direction = ord(msg[2])
                            logger.info("Set GPIO %d Dir: %s", ord(msg[1]), self.gpios[ord(msg[1])])
                            self.ack()
                        else:
                            logger.error("bad 'g' subtype: " + msg[0])
                            raise Exception
                    else:
                        if msg[0] == 'l':
                            high,mid,low = map(ord, msg[1:])
                            mask = low | mid << 8 | high << 16
                            for i in range(24):
                                self.gpios[i].level = (mask >> i) & 0x1
                            logger.info("Set GPIO level mask to: %06x", mask)
                            self.ack()
                        elif msg[0] == 'd':
                            high,mid,low = map(ord, msg[1:])
                            mask = low | mid << 8 | high << 16
                            for i in range(24):
                                self.gpios[i].direction = (mask >> i) & 0x1
                            logger.info("Set GPIO direction mask to: %06x", mask)
                            self.ack()
                        elif msg[0] == 'i':
                            high,mid,low = map(ord, msg[1:])
                            mask = low | mid << 8 | high << 16
                            for i in range(24):
                                self.gpios[i].interrupt = (mask >> i) & 0x1
                            logger.info("Set GPIO interrupt mask to: %06x", mask)
                            self.ack()
                        else:
                            logger.error("bad 'g' subtype: " + msg[0])
                            raise Exception
                elif msg_type == 'I':
                    if msg[0] == 'c':
                        logger.info("Responded to query for I2C bus speed (%d kHz)", self.i2c_speed_in_khz)
                        self.respond(struct.pack("B", self.i2c_speed_in_khz // 2))
                    elif msg[0] == 'a':
                        logger.info("Responded to query for ICE I2C mask (%02x ones %02x zeros)",
                                self.i2c_mask_ones, self.i2c_mask_zeros)
                        self.respond((self.i2c_mask_ones, self.i2c_mask_zeros))
                    else:
                        logger.error("bad 'I' subtype: " + msg[0])
                        raise Exception
                elif msg_type == 'i':
                    if msg[0] == 'c':
                        self.i2c_speed_in_khz = ord(msg[1]) * 2
                        logger.info("I2C Bus Speed set to %d kHz", self.i2c_speed_in_khz)
                        self.ack()
                    elif msg[0] == 'a':
                        self.i2c_mask_ones = ord(msg[1])
                        self.i2c_mask_zeros = ord(msg[2])
                        logger.info("ICE I2C mask set to 0x%02x ones, 0x%02x zeros",
                                self.i2c_mask_ones, self.i2c_mask_zeros)
                        self.ack()
                    else:
                        logger.error("bad 'i' subtype: " + msg[0])
                        raise Exception
                elif msg_type == 'M':
                    min_proto(2)
                    if msg[0] == 'l':
                        logger.info("Responded to query for MBus full prefix mask (%06x ones %06x zeros)",
                                self.mbus_full_prefix_ones, self.mbus_full_prefix_zeros)
                        r = struct.pack('>I', self.mbus_full_prefix_ones)[1:]
                        r += struct.pack('>I', self.mbus_full_prefix_zeros)[1:]
                        self.respond(r)
                    elif msg[0] == 's':
                        logger.info("Responded to query for MBus short prefix (%02x)",
                                self.mbus_short_prefix)
                        self.respond(struct.pack("B", self.mbus_short_prefix))
                    elif msg[0] == 'S':
                        logger.info("Responded to query for MBus snoop enabled (%d)",
                                self.mbus_snoop_enabled)
                        self.respond(struct.pack("B", self.mbus_snoop_enabled))
                    elif msg[0] == 'b':
                        logger.info("Responded to query for MBus broadcast mask (%02x ones %02x zeros)",
                                self.mbus_broadcast_mask_ones, self.mbus_broadcast_mask_zeros)
                        self.respond(struct.pack("BB",
                            self.mbus_broadcast_mask_ones,
                            self.mbus_broadcast_mask_zeros))
                    elif msg[0] == 'B':
                        logger.info("Responded to query for MBus snoop broadcast mask (%02x ones %02x zeros)",
                                self.mbus_snoop_broadcast_mask_ones, self.mbus_snoop_broadcast_mask_zeros)
                        self.respond(struct.pack("BB",
                            self.mbus_snoop_broadcast_mask_ones,
                            self.mbus_snoop_broadcast_mask_zeros))
                    elif msg[0] == 'm':
                        logger.info("Responded to query for MBus master state (%s)",
                                ("off", "on")[self.mbus_ismaster])
                        self.respond(struct.pack("B", self.mbus_ismaster))
                    elif msg[0] == 'c':
                        raise NotImplementedError("MBus clock not defined")
                    elif msg[0] == 'i':
                        logger.info("Responded to query for MBus should interrupt (%d)",
                                self.mbus_should_interrupt)
                        self.respond(struct.pack("B", self.mbus_should_interrupt))
                    elif msg[0] == 'p':
                        logger.info("Responded to query for MBus should use priority arb (%d)",
                                self.mbus_should_prio)
                        self.respond(struct.pack("B", self.mbus_should_prio))
                    elif msg[0] == 'r':
                        logger.info("Responded to query for MBus internal reset (%d)",
                                self.mbus_force_reset)
                        self.respond(struct.pack("B", self.mbus_force_reset))
                    else:
                        logger.error("bad 'M' subtype: " + msg[0])
                elif msg_type == 'm':
                    min_proto(2)
                    if msg[0] == 'l':
                        self.mbus_full_prefix_ones = ord(msg[3])
                        self.mbus_full_prefix_ones |= ord(msg[2]) << 8
                        self.mbus_full_prefix_ones |= ord(msg[1]) << 16
                        self.mbus_full_prefix_zeros = ord(msg[6])
                        self.mbus_full_prefix_zeros |= ord(msg[5]) << 8
                        self.mbus_full_prefix_zeros |= ord(msg[4]) << 16
                        logger.info("MBus full prefix mask set to ones %06x zeros %06x",
                                self.mbus_full_prefix_ones, self.mbus_full_prefix_zeros)
                        self.ack()
                    elif msg[0] == 's':
                        self.mbus_short_prefix = ord(msg[1])
                        logger.info("MBus short prefix set to %02x", self.mbus_short_prefix)
                        self.ack()
                    elif msg[0] == 'S':
                        self.mbus_snoop_enabled = ord(msg[1])
                        if self.mbus_snoop_enabled:
                            self.s_en_event.set()
                        logger.info("MBus snoop enabled set to %d", self.mbus_snoop_enabled)
                        self.ack()
                    elif msg[0] == 'b':
                        self.mbus_broadcast_mask_ones = ord(msg[1])
                        self.mbus_broadcast_mask_zeros = ord(msg[2])
                        logger.info("MBus broadcast mask set to ones %02x zeros %02x",
                                self.mbus_broadcast_mask_ones, self.mbus_broadcast_mask_zeros)
                        self.ack()
                    elif msg[0] == 'B':
                        self.mbus_snoop_broadcast_mask_ones = ord(msg[1])
                        self.mbus_snoop_broadcast_mask_zeros = ord(msg[2])
                        logger.info("MBus snoop broadcast mask set to ones %02x zeros %02x",
                                self.mbus_snoop_broadcast_mask_ones, self.mbus_snoop_broadcast_mask_zeros)
                        self.ack()
                    elif msg[0] == 'm':
                        self.mbus_ismaster = bool(ord(msg[1]))
                        logger.info("MBus master mode set " + ("off", "on")[self.mbus_ismaster])
                        self.ack()
                    elif msg[0] == 'c':
                        raise NotImplementedError("MBus clock not defined")
                    elif msg[0] == 'i':
                        self.mbus_should_interrupt = ord(msg[1])
                        logger.info("MBus should interrupt set to %d", self.mbus_should_interrupt)
                        self.ack()
                    elif msg[0] == 'p':
                        self.mbus_should_prio = ord(msg[1])
                        logger.info("MBus should use priority arbitration set to %d",
                                self.mbus_should_prio)
                        self.ack()
                    elif msg[0] == 'r':
                        self.mbus_force_reset = ord(msg[1])
                        logger.info("MBus internal reset set to %d", self.mbus_force_reset)
                        self.ack()
                    else:
                        logger.error("bad 'm' subtype: " + msg[0])
                elif msg_type == 'O':
                    if msg[0] == 'c':
                        logger.info("Responded to query for FLOW clock (%.2f Hz)", self.flow_clock_in_hz)
                        div = int(CLOCK_FREQ / self.flow_clock_in_hz)
                        resp = ''
                        if minor >= 3:
                            resp = struct.pack(">I", div)
                        else:
                            resp = struct.pack(">I", div)[1:]
                        self.respond(resp)
                    elif msg[0] == 'o':
                        if minor > 1:
                            logger.info("Responded to query for FLOW power (%s)", ('off','on')[self.flow_onoff])
                            self.respond(struct.pack("B", self.flow_onoff))
                        else:
                            logger.error("Request for protocol 0.2 command (Oo), but the")
                            logger.error("negotiated protocol was 0.1")
                    else:
                        logger.error("bad 'O' subtype: " + msg[0])
                elif msg_type == 'o':
                    if msg[0] == 'c':
                        if minor >= 3:
                            div = (ord(msg[1]) << 24) | (ord(msg[2]) << 16) | (ord(msg[3]) << 8) | ord(msg[4])
                        else:
                            div = (ord(msg[1]) << 16) | (ord(msg[2]) << 8) | ord(msg[3])
                        self.flow_clock_in_hz = CLOCK_FREQ / div
                        logger.info("Set FLOW clock to %.2f Hz", self.flow_clock_in_hz)
                        self.ack()
                    elif msg[0] == 'o':
                        min_proto(2)
                        if minor > 1:
                            self.flow_onoff = bool(ord(msg[1]))
                            logger.info("Set FLOW power to %s", ('off','on')[self.flow_onoff])
                            self.ack()
                    elif msg[0] == 'p':
                        min_proto(2)
                        ein_goc_toggle = bool(ord(msg[1]))
                        logger.info("Set GOC/EIN toggle to %s mode", ('EIN','GOC')[ein_goc_toggle])
                        self.ack()
                    else:
                        assert False
                        logger.error("bad 'o' subtype: " + msg[0])
                        assert False
                elif msg_type == 'P':
                    pwr_idx = ord(msg[1])
                    if pwr_idx not in (0,1,2):
                        logger.error("Illegal power index: %d", pwr_idx)
                        raise Exception
                    if msg[0] == 'v':
                        if pwr_idx is 0:
                            logger.info("Query 0.6V rail (vset=%d, vout=%.2f)", self.vset_0p6,
                                    (0.537 + 0.0185 * self.vset_0p6) * DEFAULT_POWER_0P6)
                            self.respond(struct.pack("BB", pwr_idx, self.vset_0p6))
                        elif pwr_idx is 1:
                            logger.info("Query 1.2V rail (vset=%d, vout=%.2f)", self.vset_1p2,
                                    (0.537 + 0.0185 * self.vset_1p2) * DEFAULT_POWER_1P2)
                            self.respond(struct.pack("BB", pwr_idx, self.vset_1p2))
                        elif pwr_idx is 2:
                            logger.info("Query VBatt rail (vset=%d, vout=%.2f)", self.vset_vbatt,
                                    (0.537 + 0.0185 * self.vset_vbatt) * DEFAULT_POWER_VBATT)
                            self.respond(struct.pack("BB", pwr_idx, self.vset_vbatt))
                    elif msg[0] == 'o':
                        if pwr_idx is 0:
                            logger.info("Query 0.6V rail (%s)", ('off','on')[self.power_0p6_on])
                            self.respond(struct.pack("B", self.power_0p6_on))
                        elif pwr_idx is 1:
                            logger.info("Query 1.2V rail (%s)", ('off','on')[self.power_1p2_on])
                            self.respond(struct.pack("B", self.power_1p2_on))
                        elif pwr_idx is 2:
                            logger.info("Query vbatt rail (%s)", ('off','on')[self.power_vbatt_on])
                            self.respond(struct.pack("B", self.power_vbatt_on))
                        elif pwr_idx is 3:
                            logger.info("Query goc rail (%s)", ('off','on')[self.power_goc_on])
                            self.respond(struct.pack("B", self.power_goc_on))
                    else:
                        logger.error("bad 'p' subtype: " + msg[0])
                        raise Exception
                elif msg_type == 'p':
                    pwr_idx = ord(msg[1])
                    if msg[0] == 'v':
                        if pwr_idx is ICE.POWER_0P6:
                            self.vset_0p6 = ord(msg[2])
                            logger.info("Set 0.6V rail to vset=%d, vout=%.2f", self.vset_0p6,
                                    (0.537 + 0.0185 * self.vset_0p6) * DEFAULT_POWER_0P6)
                        elif pwr_idx is ICE.POWER_1P2:
                            self.vset_1p2 = ord(msg[2])
                            logger.info("Set 1.2V rail to vset=%d, vout=%.2f", self.vset_1p2,
                                    (0.537 + 0.0185 * self.vset_1p2) * DEFAULT_POWER_1P2)
                        elif pwr_idx is ICE.POWER_VBATT:
                            self.vset_vbatt = ord(msg[2])
                            logger.info("Set VBatt rail to vset=%d, vout=%.2f", self.vset_vbatt,
                                    (0.537 + 0.0185 * self.vset_vbatt) * DEFAULT_POWER_VBATT)
                        else:
                            logger.error("Illegal power index: %d", pwr_idx)
                            raise Exception
                        self.ack()
                    elif msg[0] == 'o':
                        if pwr_idx is ICE.POWER_0P6:
                            self.power_0p6_on = bool(ord(msg[2]))
                            logger.info("Set 0.6V rail %s", ('off','on')[self.power_0p6_on])
                        elif pwr_idx is ICE.POWER_1P2:
                            self.power_1p2_on = bool(ord(msg[2]))
                            logger.info("Set 1.2V rail %s", ('off','on')[self.power_1p2_on])
                        elif pwr_idx is ICE.POWER_VBATT:
                            self.power_vbatt_on = bool(ord(msg[2]))
                            logger.info("Set VBatt rail %s", ('off','on')[self.power_vbatt_on])
                        elif minor >= 3 and pwr_idx is ICE.POWER_GOC:
                            self.power_goc_on = bool(ord(msg[2]))
                            logger.info("Set GOC circuit %s", ('off','on')[self.power_goc_on])
                        else:
                            logger.error("Illegal power index: %d", pwr_idx)
                            raise Exception
                        self.ack()
                    else:
                        logger.error("bad 'p' subtype: " + msg[0])
                        raise UnknownCommandException
                else:
                    logger.error("Unknown msg type: " + msg_type)
                    raise UnknownCommandException
            except UnknownCommandException:
                self.nak()
            except NameError:
                logger.error("Commands issued before version negotiation?")
                raise
            except serial.SerialException:
                logger.error("Serial Port closed on other end")
                break
            except KeyboardInterrupt:
                for th in threading.enumerate():
                    print(th)
                    traceback.print_stack(sys._current_frames()[th.ident])
                    print('------------------')
                raise

    def respond(self, msg, ack=True):
        with self.s_lock:
            if (ack):
                self.s.write(bytes((0,)))
            else:
                self.s.write(bytes((1,)))
            self.s.write(bytes((self.event,)))
            self.event += 1
            self.event %= 256
            self.s.write(bytes((len(msg),)))

            if type(msg) != bytes:
                msg = bytes(msg, 'utf-8')

            if len(msg):
                self.s.write(msg)
        logger.debug("Sent a response of length: " + str(len(msg)))

    def ack(self):
        self.respond('')

    def nak(self):
        self.respond('', ack=False)


    @staticmethod
    def get_parser():
        parser = argparse.ArgumentParser()

        parser.add_argument("-i", "--ice-version", default=4, type=int, help="Maximum ICE Version to emulate (1, 2, or 3)")
        parser.add_argument("-s", "--serial", default=_FAKE_SERIAL_SIMULATOR_ENDPOINT, help="Serial port to connect to")
        parser.add_argument("-S", "--suppress-fake-serial", action='store_false', help="Do not create a software serial tunnerl")
        parser.add_argument("--i2c-mask", default=DEFAULT_I2C_MASK, help="Address mask for fake_ice i2c address")
        parser.add_argument("-a", "--ack-all", action="store_true", help="Only supports i2c at the moment")
        parser.add_argument("-g", "--generate-messages", action="store_true", help="Generate periodic, random MBus messages")
        parser.add_argument("-r", "--replay", default=None, help="Replay a ICE snoop trace")
        parser.add_argument('-t', '--transaction', default=None, 
            help='Enter transaction mode to replay a series of ICE messages with timing')

        return parser

    def parse_cli(self):
        self.args = Simulator.get_parser().parse_args()


    def match_mask(self, val, ones, zeros):
        if self.args.ack_all:
            return True
        return ((val & ones) == ones) and ((~val & zeros) == zeros)



    def sleep(self, *args, **kwargs):
        if not hasattr(self, '_sleep'):
            try:
                os.environ['ICE_NOSLEEP']
                self._sleep = lambda x: None
            except KeyError:
                self._sleep = time.sleep
        self._sleep(*args, **kwargs)




class Gpio(object):
    GPIO_INPUT    = 0
    GPIO_OUTPUT   = 1
    GPIO_TRISTATE = 2

    def __init__(self, direction=GPIO_INPUT, level=False, interrupt=False):
        self.direction = direction
        self.level = level
        self.interrupt = interrupt

    def __str__(self):
        s = ''
        if self.direction == Gpio.GPIO_INPUT:
            s += ' IN'
        elif self.direction == Gpio.GPIO_OUTPUT:
            s += 'OUT'
        elif self.direction == Gpio.GPIO_TRISTATE:
            s += 'TRI'
        else:
            raise RuntimeError("impossible gpio direction")

        s += ' - '

        if self.level:
            s += '1'
        else:
            s += '0'

        if self.interrupt:
            s += '(int_en)'

        return s

    def __repr__(self):
        return self.__str__()

    def __setattr__(self, name, value):
        if name is 'direction':
            if value not in (Gpio.GPIO_INPUT, Gpio.GPIO_OUTPUT, Gpio.GPIO_TRISTATE):
                raise ValueError("Attempt to set illegal direction {}".format(value))
        if name is 'level':
            if value not in (True, False):
                raise ValueError("GPIO level must be true or false. Got {}".format(value))
        if name is 'interrupt':
            if value not in (True, False):
                raise ValueError("GPIO interrupt must be true or false. Got {}".format(value))
        object.__setattr__(self, name, value)


_socat_time = str(datetime.datetime.now())
_socat_fpre = os.path.join(tempfile.gettempdir(), _socat_time + '-')
_socat_proc = None
_socat_devnull = None

if platform.system() == 'Darwin':
    # Well-intentioned private temp directories are annoying in this case
    tempfile.tempdir = '/tmp'
_FAKE_SERIAL_CONNECTTO_ENDPOINT = os.path.join(tempfile.gettempdir(), 'm3_ice_com1')
_FAKE_SERIAL_SIMULATOR_ENDPOINT = os.path.join(tempfile.gettempdir(), 'm3_ice_com2')

def destroy_fake_serial():
    global _socat_proc
    global _socat_devnull

    if _socat_proc:
        _socat_proc.terminate()
        _socat_proc.wait()
        _socat_proc = None
    if _socat_devnull:
        _socat_devnull.close()
        _socat_devnull = None

def create_fake_serial(
        endpoint1=_FAKE_SERIAL_CONNECTTO_ENDPOINT,
        endpoint2=_FAKE_SERIAL_SIMULATOR_ENDPOINT,
        ):
    global _socat_fpre
    global _socat_proc
    global _socat_devnull

    _socat_devnull = open(os.devnull, 'w')
    _socat_proc = subprocess.Popen(
                "socat -x pty,link={},raw,echo=0 pty,link={},raw,echo=0".format(endpoint1, endpoint2),
                stdout=open(_socat_fpre + 'socat-stdout', 'w'),
                stderr=open(_socat_fpre + 'socat-stderr', 'w'),
                shell=True,
                )

    # Hack, b/c socat doesn't exit but do need to wait for pipe to be set up
    limit = time.time() + 5
    while not (os.path.exists(endpoint1) and os.path.exists(endpoint2)):
        time.sleep(.1)
        if time.time() > limit:
            _socat_proc.kill()
            for l in open(_socat_fpre + 'socat-stdout'):
                logger.debug(l)
            for l in open(_socat_fpre + 'socat-stderr'):
                logger.debug(l)
            raise NotImplementedError("socat endpoint never appeared?")

    logger.debug("Fake serial bridge created.")

    atexit.register(destroy_fake_serial)


def cmd():
    try:
        Simulator().run()
    except KeyboardInterrupt:
        logger.info("Caught quit request. Shutting down.")

if __name__ == '__main__':
    Simulator().run()





try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from m3 import __version__

config = {
        'name': 'm3',
        'packages': ['m3'],
        'version': __version__,
        'author': 'Pat Pannuto',
        'author_email': 'pat.pannuto@gmail.com',
        'url': 'https://github.com/mbus/m3-python',


        'description': 'Support tools for chips from the M3 ecosystem',
        'long_description': '''
M3 Python ICE
=============

This Python library provides tools for interfacing with the ICE debug
and development board, part of the M3 ecosystem.

In addition, several high-level tools such as a programmer, message
injector, and snooper are included with this package.''',
        'classifiers': [
            "Development Status :: 3 - Alpha",
            "Environment :: Console",
            "Intended Audience :: Developers",
            "Intended Audience :: End Users/Desktop",
            "License :: OSI Approved :: MIT License",
            "Natural Language :: English",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 2.7",
            #"Programming Language :: Python :: 3",
            #"Programming Language :: Python :: 3.3",
            #"Programming Language :: Python :: 3.4",
            #"Programming Language :: Python :: 3.5",
            #"Programming Language :: Python :: Implementation :: CPython",
            #"Programming Language :: Python :: Implementation :: PyPy",
            "Topic :: Software Development :: Embedded Systems",
            ],

        'install_requires': [
            'future',
            'nose',
            'pyserial',
            ],
        'entry_points': {
            'console_scripts': [
                'm3_ice_simulator = m3.ice_simulator:cmd',
                'm3_ice           = m3.m3_ice:cmd',
                ],
            },
        }

setup(**config)



#!/usr/bin/env python

################################################################################

# Coerce Py2k to act more like Py3k
from __future__ import (absolute_import, division, print_function, unicode_literals)
from builtins import (
        ascii, bytes, chr, dict, filter, hex, input, int, isinstance, list, map,
        next, object, oct, open, pow, range, round, str, super, zip,
        )

import binascii
from copy import copy
from copy import deepcopy
import errno
import functools
import socket
import struct
import sys
import time
import os

from . import m3_logging
logger = m3_logging.getLogger(__name__)

try:
    import threading
    import queue as Queue
except ImportError:
    logger.warn("Your python installation does not support threads.")
    logger.warn("")
    logger.warn("Please install a version of python that supports threading.")
    raise

try:
    import serial
except ImportError:
    logger.warn("You do not have the pyserial library installed.")
    logger.warn("")
    logger.warn("For debian-based systems (e.g. Ubuntu):")
    logger.warn("\tsudo apt-get install pyserial")
    logger.warn("For rpm-based systems (e.g. Red Hat):")
    logger.warn("\tsudo yum install pyserial")
    logger.warn("For more installation instructions + see:")
    logger.warn("\thttp://pyserial.sourceforge.net/pyserial.html#installation")
    raise

################################################################################

class ICE(object):
    VERSIONS = ((0,1),(0,2),(0,3),(0,4),(0,5))
    ONEYEAR = 365 * 24 * 60 * 60

    class ICE_Error(Exception):
        '''
        A base class for all exceptions raised by this module
        '''
        pass

    class NotConnectedError(ICE_Error):
        '''
        A method was called that requires ICE to be connected when it is not.
        '''
        pass

    class FormatError(ICE_Error):
        '''
        Something in the ICE protocol communicating with the board went wrong.

        This error should never be encountered in normal use.
        '''
        pass

    class ParameterError(ICE_Error):
        '''
        An illegal parameter was passed.

        This may be raised by the ICE library if it can determine in advance
        that the reqeust is illegal (e.g. out of range), or by the ICE board if
        the board rejects the desired setting (e.g. not configurable)
        '''

    class NAK_Error(ICE_Error):
        '''
        Raised when an unexpected NAK is returned
        '''
        pass

    class VersionError(ICE_Error):
        '''
        A method was called that attached ICE version does not support.
        '''
        def __init__(self, required_version, current_version):
            self.required_version = required_version
            self.current_version = current_version
            super(ICE.VersionError, self).__init__()

    class CapabilityError(ICE_Error):
        '''
        A method was called that the attached ICE board does not have hardware
        frontend for.
        '''
        def __init__(self, required_capability, capabilities):
            self.required_capability = required_capability
            self.capabilities = capabilities
            super(ICE.CapabilityError, self).__init__(
                    "Missing required capability `{}`. Available capabilities are `{}`".\
                    format(required_capability, capabilities)
                    )

    class TimeoutError(ICE_Error):
        ''' 
        A method was called that attempted to read more bytes than the Serial
        interface had available within the timeout window
        '''
        def __init__(self, _timeout, _partial_data):
            self.timeout = _timeout
            self.partial_data = _partial_data
            super(ICE.TimeoutError, self).__init__()

    ## Support decorators:
    def min_proto_version(version):
        '''
        Decorator for library calls that verifies the requested call is
        supported by the protocol version negotiated by the current ICE board.
        '''
        def wrapped_fn_factory(fn_being_decorated):
            @functools.wraps(fn_being_decorated)
            def wrapped_fn(self, *args, **kwargs):
                if not hasattr(self, "minor"):
                    raise self.NotConnectedError("ICE must be connected first ({})".format(fn_being_decorated))
                major, minor = map(int, version.split('.'))
                if major != 0:
                    raise self.ICE_Error("Major version bump?")
                if self.minor < minor:
                    raise self.VersionError(minor, self.minor)
                return fn_being_decorated(self, *args, **kwargs)
            return wrapped_fn
        return wrapped_fn_factory

    def max_proto_version(version):
        '''
        Decorator for library calls that verifies the requested call is
        supported by the protocol version negotiated by the current ICE board.
        '''
        def wrapped_fn_factory(fn_being_decorated):
            @functools.wraps(fn_being_decorated)
            def wrapped_fn(self, *args, **kwargs):
                if not hasattr(self, "minor"):
                    raise self.ICE_Error("ICE must be connected first")
                major, minor = map(int, version.split('.'))
                if major != 0:
                    raise self.ICE_Error("Major version bump?")
                if self.minor > minor:
                    raise self.VersionError(minor, self.minor)
                return fn_being_decorated(self, *args, **kwargs)
            return wrapped_fn
        return wrapped_fn_factory

    def capability(cap):
        '''
        Decorator for library calls that verifies the requested call is
        supported by the capabilities reported by the current ICE board.
        '''
        def wrapped_fn_factory(fn_being_decorated):
            @functools.wraps(fn_being_decorated)
            def wrapped_fn(self, *args, **kwargs):
                try:
                    if cap not in self.capabilities:
                        raise self.CapabilityError(cap, self.capabilities)
                except AttributeError:
                    if 'ice_query_capabilities' not in fn_being_decorated.__name__:
                        if self.minor != 1:
                            logger.error("Version decorator must precede capability")
                            raise
                return fn_being_decorated(self, *args, **kwargs)
            return wrapped_fn
        return wrapped_fn_factory

    def __init__(self):
        '''
        An ICE object.

        Most methods are not usuable until connect() has been called.
        '''

        self.event_id = 0
        self.last_event_id = -1
        self.sync_queue = Queue.Queue(1)

        self.msg_handler = {}
        self.d_lock = threading.Lock()
        self.d_frag = bytes()
        self.msg_handler['d'] = self.d_defragger
        self.b_lock = threading.Lock()
        self.b_frag = bytes()
        self.msg_handler['b'] = self.b_defragger
        self.B_lock = threading.Lock()
        self.B_frag = bytes()
        self.msg_handler['B'] = self.B_defragger
        self.B_formatter_success_only = False
        self.B_formatter_control_bits = False
        self.msg_handler['B+'] = self.B_formatter
        self.msg_handler['b+'] = self.b_formatter

        self.msg_handler['e'] = self.e_handler

        self.goc_ein_toggle = -1

        # Set initial, minimal capability set
        self.capabilities = 'VvXx'

    def find_baud(self, serial_device):
       
        # we're only trying these
        baudrates = [  115200, 2000000] 
        version_request = binascii.unhexlify('560000')
        found = False
        
        with serial.Serial(serial_device, baudrates[0], 
                    timeout=0.05 ) as tmpSerial:

            if not tmpSerial.isOpen():
                raise self.ICE_Error("Failed to connect to temporary serial device")

            for baudrate in baudrates:

                logger.debug('Trying baudrate: ' + str(baudrate))
                try:
                    tmpSerial.baudrate = baudrate
                except IOError: 
                    logger.debug("Error changing baudrate, assuming socat port")
                    found = baudrate
                    break
                       
                # send a version request and see what happens
                tmpSerial.write(version_request)
                rxBytes = tmpSerial.read(5) #see if this times out
                if len(rxBytes) != 0: 
                    found = baudrate
                    break
            
        if not found:
            raise Exception("Unable to determine baudrate!")

        logger.debug ("Found Baudrate: " + str(baudrate) )
        return found 

    def connect(self, serial_device, baudrate=115200):
        '''
        Opens a connection to the ICE board.

        The ICE object configuration (e.g. message handlers) cannot be safely
        changed after this method is invoked.
        '''

        #500ms timeout for serial to help catch runaway packets
        # cygwin cannot support 5ms or 50ms timeouts 
        # m3_ice_sim doesn't support baudrate
        try:
            self.dev = serial.Serial(serial_device, baudrate, timeout=0.5)
        except IOError:
            logger.warn("Skipping baudrate?")
            self.dev = serial.Serial(serial_device, timeout=0.5)

        if self.dev.isOpen():
            logger.info("Connected to serial device at " + self.dev.portstr + 
                " at " + str(baudrate) + " baud")
        else:
            raise self.ICE_Error("Failed to connect to serial device")

        self.communicator_stop_request = threading.Event()
        self.communicator_stop_response = threading.Event()
        self.comm_thread = threading.Thread(target=self.communicator)
        self.comm_thread.daemon = True
        self.comm_thread.start()

        self.negotiate_version()

        if self.minor == 2:
            # V2 ICE sets GOC on by default, which is annoying. Correct that.
            self.goc_set_onoff(False)

    def is_connected(self):
        return hasattr(self, 'dev')

    def destroy(self):
        if hasattr(self, 'dev'):
            self.communicator_stop_request.set()
            self.communicator_stop_response.wait()
            self.dev.close()
            logger.info("Connection to " + self.dev.portstr + " closed.")
            del(self.dev)

    def spawn_handler(self, msg_type, event_id, length, msg):
        try:
            handler = self.msg_handler[msg_type]
            #t = threading.Thread(target=self.msg_handler[msg_type],
            #        args=(msg_type, event_id, length, msg))
            #t.daemon = True
            #t.start()
        except KeyError:
            if msg_type not in self.capabilities:
                logger.warn("Synchronization lost. Likely causes:")
                logger.warn("  - you are reconnecting to an ICE that was previously snooping")
                logger.warn("  - your computer can't keep up with the rate of messages ICE sends")
                logger.warn("  - some transient serial error occurred (not impossible at 3 MBaud)")
                logger.warn("This library will try to get back on track, but if")
                logger.warn("this message keeps printing, you'll need to hit the")
                logger.warn("reset button on the ICE board")
                # The idea here is to read as much as available in the serial
                # buffer, throwing it away, and count on the gaps between ICE
                # messages to get things back on track. A bit ugly, but I'm not
                # sure I know of a better solution :/
                self.dev.read()
            try:
                logger.warn("WARNING: No handler registered for message type: " +
                        str(msg_type))
                logger.warn("Known Types:")
                for t,f in self.msg_handler.iteritems():
                    logger.warn("%s\t%s" % (t, str(f)))
                logger.warn("         Dropping packet:")
                logger.warn("")
                logger.warn("    Type: %s" % (msg_type))
                logger.warn("Event ID: %d" % (event_id))
                logger.warn("  Length: %d" % (length))
                logger.warn(" Message:" + msg.encode('hex'))
            except Exception as e:
                logger.warn("Unhandled exception trying to report unknown message.")
                logger.warn(str(e))
                logger.warn("Suppressed.")
            return
        try:
            handler(msg_type, event_id, length, msg)
        except self.NotConnectedError:
            # The ICE board can send async messages before the library is set up
            # to receive them, silently drop
            logger.debug("Received message from ICE before connection established, dropping.")
            try:
                logger.debug("         Dropping packet:")
                logger.debug("")
                logger.debug("    Type: %s" % (msg_type))
                logger.debug("Event ID: %d" % (event_id))
                logger.debug("  Length: %d" % (length))
                logger.debug(" Message:" + msg.encode('hex'))
            except Exception as e:
                logger.debug("Unhandled exception trying to report unknown message.")
                logger.debug(str(e))
                logger.debug("Suppressed.")

    def useful_read(self, length, check_timeout = False):
        rxBuf = b''
        while len(rxBuf) < length:
            rx = self.dev.read(length - len(rxBuf))

            if check_timeout and len(rx) == 0: #timeout occured
                raise self.TimeoutError(self.dev.timeout, rxBuf)

            else: # add to the buffer
                rxBuf += rx
    
        assert len(rxBuf) == length
        logger.debug('Raw Read: ' + binascii.hexlify(rxBuf) )
        return rxBuf 
       
    def communicator(self):
        while not self.communicator_stop_request.isSet():
            try:
                # Read has a timeout of .1 s. Polling is the easiest way to
                # do x-platform cancellation
                msg_type, event_id, length = self.useful_read(3)
            except ValueError:
                continue
            except (serial.SerialException, OSError):
                break
            msg_type = ord(msg_type)
            event_id = ord(event_id)
            length = ord(length)
            #print("Got msg type", msg_type, chr(msg_type), length)
            try:
                msg = self.useful_read(length, check_timeout = True)
            except self.TimeoutError:
                logger.warn("Timeout error occured, skipping rest of packet!")
                continue
            #print(msg.encode('hex'))

            if event_id == self.last_event_id:
                logger.warn("WARNING: Duplicate event_id! THIS IS A BUG [somewhere]!!")
                logger.warn("         Dropping packet:")
                logger.warn("")
                logger.warn("    Type: %d" % (msg_type))
                logger.warn("Event ID: %d" % (event_id))
                logger.warn("  Length: %d" % (length))
                logger.warn(" Message:" + msg.encode('hex'))
            else:
                self.last_event_id = event_id

            if msg_type in (0,1):
                # Ack / Nack response from a synchronous message
                try:
                    if msg_type == 0:
                        logger.debug("Got an ACK packet. Event: " + str(event_id))
                    else:
                        logger.info("Got a NAK packet. Event:" + str(event_id))
                    self.sync_queue.put((msg_type, msg))
                except Queue.Full:
                    logger.warn("WARNING: Synchronization lost. Unsolicited ACK/NAK.")
                    logger.warn("         Dropping packet:")
                    logger.warn("")
                    logger.warn("    Type: %s" % (["ACK","NAK"][msg_type]))
                    logger.warn("Event ID: %d" % (event_id))
                    logger.warn("  Length: %d" % (length))
                    logger.warn(" Message:" + msg.encode('hex'))
            else:
                msg_type = chr(msg_type)
                logger.debug("Got an async message of type: " + msg_type)
                self.spawn_handler(msg_type, event_id, length, msg)
        self.communicator_stop_response.set()
        if hasattr(self, 'on_disconnect'):
            self.on_disconnect()

    def string_to_masks(self, mask_string):
        ones = 0
        zeros = 0
        mask_string = mask_string.replace(' ','')
        idx = len(mask_string)
        for c in mask_string:
            idx -= 1
            if c == '1':
                ones |= (1 << idx)
            elif c == '0':
                zeros |= (1 << idx)
            elif c in ('x', 'X'):
                continue
            else:
                raise self.FormatError("Illegal character: >>>" + c + "<<<")
        return ones,zeros

    def masks_to_strings(self, ones, zeros, length):
        s = ''
        for l in range(length):
            o = bool(ones & (1 << l))
            z = bool(zeros & (1 << l))
            if o and z:
                raise self.FormatError("masks_to_strings has req 1 and req 0." +
                        "ones {} zeros {} length {} l {}".format(ones, zeros, length, l))
            if o:
                s = '1' + s
            elif z:
                s = '0' + s
            else:
                s = 'x' + s
        return s

    def d_defragger(self, msg_type, event_id, length, msg):
        '''
        Helper function to defragment 'd' type I2C messages before forwarding.

        This helper is installed by default for 'd' messages. It will attempt to
        call a helper registered under the name 'd+' when a complete message has
        been received. The message will be assigned the event id of the last
        received fragment.

        It may be safely overridden.
        '''
        with self.d_lock:
            assert msg_type == 'd'
            self.d_frag += msg
            # XXX: Make version dependent
            if length != 255:
                sys.stdout.flush()
                logger.debug("Got a complete I2C transaction of length %d bytes. Forwarding..." % (len(self.d_frag)))
                sys.stdout.flush()
                self.spawn_handler('d+', event_id, len(self.d_frag), \
                        deepcopy(self.d_frag))
                self.d_frag = bytes()
            else:
                logger.debug("Got an I2C fragment... thus far %d bytes received:" % (len(self.d_frag)))

    @min_proto_version("0.2")
    def b_defragger(self, msg_type, event_id, length, msg):
        '''
        Helper function to defragment 'b' type MBus messages before forwarding.

        This helper is installed by default for 'b' messages. It will attempt to
        call a helper registered under the name 'b+' when a complete message has
        been received. The message will be assigned the event id of the last
        received fragment.

        It may be safely overridden.
        '''
        with self.b_lock:
            logger.debug("\tmsg_type: %s, event_id: %s, length: %s, msg: %s"
                    % (msg_type, event_id, length, repr(msg)))
            assert msg_type == 'b'
            self.b_frag += msg
            # XXX: Make version dependent
            if length != 255:
                logger.debug("Got a complete MBus message of length %d bytes. Forwarding..." % (len(self.b_frag)))
                self.spawn_handler('b+', event_id, len(self.b_frag), \
                        deepcopy(self.b_frag))
                self.b_frag = bytes()
            else:
                logger.debug("Got a MBus fragment... thus far %d bytes received:" % (len(self.b_frag)))

    @min_proto_version("0.2")
    def B_defragger(self, msg_type, event_id, length, msg):
        '''
        Helper function to defragment 'B' type snooped MBus messages before forwarding.

        This helper is installed by default for 'B' messages. It will attempt to
        call a helper registered under the name 'B+' when a complete message has
        been received. The message will be assigned the event id of the last
        received fragment.

        It may be safely overridden.
        '''
        with self.B_lock:
            assert msg_type == 'B'
            self.B_frag += msg
            # XXX: Make version dependent
            if length != 255:
                sys.stdout.flush()
                logger.debug("Got a complete snooped MBus message. Length %d bytes. Forwarding..." % (len(self.B_frag)))
                sys.stdout.flush()
                self.spawn_handler('B+', event_id, len(self.B_frag), \
                        deepcopy(self.B_frag))
                self.B_frag = bytes()
            else:
                logger.debug("Got a snoop MBus fragment... thus far %d bytes received:" % (len(self.B_frag)))

    @min_proto_version("0.2")
    def B_formatter(self, msg_type, event_id, length, msg):
        return self.common_bB_formatter(msg_type, event_id, length, msg, 'B++')

    @min_proto_version("0.2")
    def b_formatter(self, msg_type, event_id, length, msg):
        return self.common_bB_formatter(msg_type, event_id, length, msg, 'b++')

    @min_proto_version("0.2")
    def common_bB_formatter(self, msg_type, event_id, length, msg, b_type):
        '''
        Helper function that parses 'B+' snooped MBus messages before forwarding.

        This helper is installed by default for 'B+' messages. It will attempt
        to call a helper registered under the name 'B++' when a complete message
        has been received. B++ messages do not have the standard signature,
        instead they expect a callback of the form:

            Bpp_callback(address, data)
              or
            Bpp_callback(address, data, control_bit_0, control_bit_1)

        The member variable "B_formatter_success_only" (default False) controls
        whether all messages are forwarded or only messages that were ACK'd.

        The member variable "B_formatter_control_bits" (default False) controls
        whether the control bits are sent.

        This function may be safely overridden.
        '''
        addr = msg[0:4]
        data = msg[4:-1]
        cb = ord(msg[-1:])
        # status_bits <= `SD status_bits | {4'b0000, mbus_rxfail, mbus_rxbcast, ice_export_control_bits};
        cb0 = bool(cb & 0x1)
        cb1 = bool(cb & 0x2)
        success = cb0 & (~cb1) # XXX Something is wrong here [also fix default]
        try:
            handler = self.msg_handler[b_type]
        except KeyError:
            logger.warn("All registered handlers: {}".format(self.msg_handler))
            logger.warn("Looking up key >>{}<<".format(b_type))
            try:
                logger.warn("No handler registered for B++ (formatted, snooped MBus) messages")
                logger.warn("Dropping message:")
                logger.warn("\taddr: " + binascii.hexlify(addr))
                logger.warn("\tdata: " + binascii.hexlify(data))
                logger.warn("\tstat: " + binascii.hexlify(cb))
                logger.warn("")
            except Exception as e:
                logger.warn("Unhandled exception trying to report missing B++ handler.")
                logger.warn(str(e))
                logger.warn("Suppressed.")
            return
        if self.B_formatter_control_bits:
            handler(addr, data, cb0, cb1)
        else:
            handler(addr, data)

    @min_proto_version("0.4")
    def e_handler(self, msg_type, event_id, length, msg):
        '''
        Helper function for the 'e' command, exit
        '''
        logger.info('Caught "e" command, shutting down')
        #sys.exit(0)
        os._exit( int(binascii.hexlify(msg) ,16) )

    def send_message(self, msg_type, msg='', length=None):
        if type(msg_type) != bytes:
            msg_type = bytes(msg_type, 'utf-8')

        if len(msg_type) != 1:
            raise self.FormatError("msg_type must be exactly 1 byte")

        if type(msg) != bytes:
            msg = bytes(msg, 'utf-8')

        if len(msg) > 255:
            raise self.FormatError("msg too long. Maximum msg is 255 bytes")

        if length is None:
            length = len(msg)

        buf = struct.pack("BBB", ord(msg_type), self.event_id, length)
        self.event_id = (self.event_id + 1) % 256
        logger.debug('Sending %s', binascii.hexlify(buf+msg))
        self.dev.write(buf + msg)

        # Ugly hack so python allows keyboard interrupts
        return self.sync_queue.get(True, self.ONEYEAR)

    def send_message_until_acked(self, msg_type, msg='', length=None, tries=5):
        while tries:
            ack, msg = self.send_message(msg_type, msg, length)
            if ack == 0:
                return msg
            tries -= 1

        raise self.NAK_Error

    def negotiate_version(self):
        '''
        Establish communication with an ICE board.

        This function is called automatically by __init__ and should not be
        called directly. For ICE versions >= 0.2, this function automatically
        calls ice_query_capabilities and sets up the ICE library appropriately.
        '''
        logger.info("This library supports versions...")
        for major, minor in ICE.VERSIONS:
            logger.info("\t%d.%d" % (major, minor))

        logger.debug("Sending version probe")
        resp = self.send_message_until_acked('V')
        if (len(resp) is 0) or (len(resp) % 2):
            raise self.FormatError("Version response: " + resp)

        logger.info("This ICE board supports versions...")
        self.major = None
        self.minor = None
        while len(resp) > 0:
            major, minor = struct.unpack("BB", resp[:2])
            resp = resp[2:]
            if self.major is None and (major, minor) in ICE.VERSIONS:
                self.major = major
                self.minor = minor
                logger.info("\t%d.%d **Chosen version" % (major, minor))
            else:
                logger.info("\t%d.%d" % (major, minor))

        if self.major is None:
            logger.warn("No versions in common. Version negotiation failed.")
            raise self.ICE_Error

        if self.major != 0:
            logger.error("Major version number bump. Need to re-examine python versioning")
            raise self.ICE_Error

        self.send_message_until_acked('v', struct.pack("BB", self.major, self.minor))

        if self.minor >= 2:
            logger.debug("ICE version supports capabilities, querying")
            self.capabilities = 'VvXx?'
            self.ice_query_capabilities()
            logger.debug("Capabilities: " + self.capabilities)
        else:
            self.capabilities = 'VvXxdIifOoGgPp'
            logger.debug("Version 0.1 does not have capability support, skipping")

    def min_version(self, required_version):
        if required_version > 1:
            logger.error("Need to fix this versioning system. Major version number bumped")
            raise self.ICE_Error
        required_version = int(required_version * 10)
        try:
            if self.minor < required_version:
                raise self.VersionError(required_version, self.minor)
        except AttributeError:
            logger.error("Attempt to call method before version negotiation?")
            raise

    def _fragment_sender(self, msg_type, msg):
        '''
        Internal. (helper for {i2c,goc,ein,mbus}_send)
        '''
        # XXX: Make version dependent?
        FRAG_SIZE = 255
        retry = True 

        sent = 0
        logger.debug("Sending %d byte message (in %d byte fragments)" % \
                                                    (len(msg), FRAG_SIZE))
        while len(msg) >= FRAG_SIZE:
            ack,resp = self.send_message(msg_type, msg[0:FRAG_SIZE])
            if ack == 1: # (NAK)
                if len(resp) == 0 and retry:
                    logger.warning("ICE NAK'd request to send with no length "\
                                    "sent field, assuming 0 and retrying")
                    retry = False
                else: return sent + ord(resp)
            else: retry = True

            msg = msg[FRAG_SIZE:]
            sent += FRAG_SIZE
            logger.debug("\tSent %d byte s, %d remaining" % (sent, len(msg)))

        logger.debug("Sending last message fragment, %d bytes long" % \
                                                            (len(msg)))
        while True:
            ack,resp = self.send_message(msg_type, msg)
            if ack == 1:
                if len(resp) == 0 and retry:
                    logger.warning("ICE NAK'd request to send with no length "\
                                    "sent field, assuming 0 and retrying")
                    retry = False
                else: return sent + ord(resp)
            else: break # no more while loop
        sent += len(msg)
        return sent

    ## QUERY / CONFIGURE ICE ##
    @min_proto_version("0.2")
    @capability('?')
    def ice_query_capabilities(self):
        '''
        Queries ICE board for available hardware frontends.

        The ICE library will be configured to raise an ICE.CapabilityError
        if a request that is unsupported by this hardware is requested.

        This interface is very raw and needs to be wrapped in something more
        user-friendly and library-esque. It currently returns the raw array of
        characters from the ICE board, which requires the caller to know the
        ICE protocol.
        '''
        resp = self.send_message_until_acked('?', struct.pack("B", ord('?')))
        self.capabilities = resp
        return resp

    @min_proto_version("0.2")
    @capability('?')
    def ice_get_baudrate(self):
        '''
        Gets the current baud rate of the ICE bridge in Hz.

        XXX: Returns the ideal value, not the exact speed. Not sure which is
        more correct / more useful.
        '''
        self.min_version(0.2)
        resp = self.send_message_until_acked('?', struct.pack("B", ord('b')))
        div = struct.unpack("!H", resp)[0]

        if div == 0x00AE:
            return 1152200
        elif div == 0x0007:
            return 3000000
        else:
            raise self.FormatError("Unknown baud divider?")

    @min_proto_version("0.2")
    @capability('_')
    def ice_set_baudrate(self, div, baudrate):
        '''
        Sets a new baud rate for the ICE bridge.

        Internal. This function is not meant to be called directly.
        '''
        self.min_version(0.2)
        self.send_message_until_acked('_', struct.pack("!BH", ord('b'), div))
        try:
            self.dev.baudrate = baudrate
        except IOError as e:
            if e.errno == 25:
                logger.warn("Failed to set baud rate (if socat, ignore)")
            else:
                raise

    @min_proto_version("0.2")
    @capability('_')
    def ice_set_baudrate_to_115200(self):
        self.ice_set_baudrate(0x00AE, 115200)

    @min_proto_version("0.2")
    @capability('_')
    def ice_set_baudrate_to_230400(self):
        self.ice_set_baudrate(0x00AE//2, 115200*2)

    @min_proto_version("0.2")
    @capability('_')
    def ice_set_baudrate_to_460800(self):
        self.ice_set_baudrate(0x00AE//4, 115200*4)

    @min_proto_version("0.2")
    @capability('_')
    def ice_set_baudrate_to_921600(self):
        self.ice_set_baudrate(0x00AE//8, 115200*8)

    @min_proto_version("0.2")
    @capability('_')
    def ice_set_baudrate_to_1843200(self):
        self.ice_set_baudrate(0x00AE//16, 115200*16)

    @min_proto_version("0.2")
    @capability('_')
    def ice_set_baudrate_to_2000000(self):
        self.ice_set_baudrate(0x000A, 2000000)

    @min_proto_version("0.2")
    @capability('_')
    def ice_set_baudrate_to_3_megabaud(self):
        self.ice_set_baudrate(0x0007, 3000000)


    ## GOC VS EIN HANDLING ##
    def get_goc_enabled(self):
        return self.goc_ein_toggle > 0

    def get_ein_enabled(self):
        return self.goc_ein_toggle == 0

    def set_goc_ein(self, goc=0, ein=0, goc_ir=0, restore_clock_freq=True):
        if ( (goc>0) + (ein>0) + (goc_ir>0) ) > 1:
            raise self.ICE_Error("Internal consistency goc vs ein failure")

        if self.minor == 1:
            if goc == 1:
                return
            else:
                raise self.ICE_Error("Attempt to call set_goc_ein for ein with protocol version 1")

        if ein:
            # Set to EIN mode
            if self.goc_ein_toggle == 0:
                # Already in ein mode, nothing to do
                return
            if self.goc_ein_toggle >= 1:
                # If we were set to GOC mode, capture the clock frequency
                self.goc_freq_divisor = self.goc_ein_get_freq_divisor()
            if restore_clock_freq:
                try:
                    self.goc_ein_set_freq_divisor(self.ein_freq_divisor)
                    logger.debug("Restored previous EIN clock frequency")
                except AttributeError:
                    self.goc_ein_set_freq_divisor(self.EIN_DEFAULT_DIVISOR)
                    logger.debug("Set EIN to default clock frequency")
            self.send_message_until_acked('o', struct.pack("BB", ord('p'), 0))
            self.goc_ein_toggle = 0
            logger.debug("Set goc/ein toggle to ein")
        elif (goc >= 1) or (goc_ir >= 1):
            assert( goc != goc_ir )

            # 1: visiable LED, 3: infrared LED
            goc_ctrl_byte = 1 if (goc>=1) else 3

            # Set to GOC mode
            if self.goc_ein_toggle == goc_ctrl_byte:
                return
            if self.goc_ein_toggle == 0:
                self.ein_freq_divisor = self.goc_ein_get_freq_divisor()
            if restore_clock_freq:
                try:
                    self.goc_ein_set_freq_divisor(self.goc_freq_divisor)
                    logger.debug("Restored previous GOC clock frequency")
                except AttributeError:
                    self.goc_ein_set_freq_divisor(self._goc_freq_in_hz_to_divisor(self.GOC_SPEED_DEFAULT_HZ))
                    logger.debug("Set GOC to default clock frequency")
            self.send_message_until_acked('o', 
                    struct.pack("BB", ord('p'), goc_ctrl_byte))
            self.goc_ein_toggle = goc_ctrl_byte
            logger.debug("Set goc/ein toggle to goc")
        else: raise Exception('Unsupported GOC/EIN mode')

    @max_proto_version("0.2")
    def goc_ein_get_freq_divisor_max_0_2(self):
        resp = self.send_message_until_acked('O', struct.pack("B", ord('c')))
        if len(resp) != 3:
            raise self.FormatError("Wrong response length from `Oc': " + str(resp))
        setting = struct.unpack("!I", "\x00"+resp)[0]
        return setting

    @min_proto_version("0.3")
    def goc_ein_get_freq_divisor_min_0_3(self):
        resp = self.send_message_until_acked('O', struct.pack("B", ord('c')))
        if len(resp) != 4:
            raise self.FormatError("Wrong response length from `Oc': " + str(resp))
        setting = struct.unpack("!I", resp)[0]
        logger.debug('got divisor value {}'.format(setting))
        return setting

    def goc_ein_get_freq_divisor(self):
        if self.minor > 2:
            return self.goc_ein_get_freq_divisor_min_0_3()
        else:
            return self.goc_ein_get_freq_divisor_max_0_2()

    @max_proto_version("0.2")
    def goc_ein_set_freq_divisor_max_0_2(self, divisor):
        packed = struct.pack("!I", divisor)
        if packed[0] != '\x00':
            raise self.ParameterError("Out of range.")
        msg = struct.pack("B", ord('c')) + packed[1:]
        self.send_message_until_acked('o', msg)

    @min_proto_version("0.3")
    def goc_ein_set_freq_divisor_min_0_3(self, divisor):
        logger.debug('set divisor to {}'.format(divisor))
        packed = struct.pack("!I", divisor)
        msg = struct.pack("B", ord('c')) + packed
        self.send_message_until_acked('o', msg)

    def goc_ein_set_freq_divisor(self, divisor):
        if self.minor > 2:
            return self.goc_ein_set_freq_divisor_min_0_3(divisor)
        else:
            return self.goc_ein_set_freq_divisor_max_0_2(divisor)

    ## GOC ##
    GOC_SPEED_DEFAULT_HZ = .625

    def _goc_display_delay(self, msg, event):
        try:
            freq = self.goc_freq
        except AttributeError:
            freq = ICE.GOC_SPEED_DEFAULT_HZ

        num_bits = len(msg) * 8
        t = num_bits / freq
        logger.info("Sleeping for %f seconds while it blinks..." % (t))
        while (t > 1):
            sys.stdout.write("\r\t\t\t\t\t\t")
            sys.stdout.write("\r\t%f remaining..." % (t))
            sys.stdout.flush()
            t -= 1
            if event.is_set():
                return
            time.sleep(1)
        time.sleep(t)

    @min_proto_version("0.1")
    @capability('f')
    def goc_send(self, msg, show_progress=True, encoding='PWM'):
        '''
        Blinks a message via GOC.

        Takes a raw byte stream (e.g. binascii.unhexlify("aa")).
        Returns the number of bytes actually sent.

        Long messages may be fragmented between the ICE library and the ICE
        FPGA. These fragments will be combined on the ICE board, and given the
        significantly lower bandwidth of the GOC interface, there should be no
        interruption in message transmission.
        '''
        if not self.get_goc_enabled():
            self.set_goc_ein(goc=1)

        encoding_to_msg_type = {
                'PWM': 'f',
                'manchester': 'n',
                }

        if show_progress:
            e = threading.Event()
            t = threading.Thread(target=self._goc_display_delay, args=(msg,e))
            t.daemon = True
            t.start()
            ret = self._fragment_sender(encoding_to_msg_type[encoding], msg)
            e.set()
            t.join()
        else:
            ret = self._fragment_sender(encoding_to_msg_type[encoding], msg)
        return ret

    @min_proto_version("0.1")
    @capability('O')
    def goc_get_frequency(self):
        '''
        Gets the GOC frequency.
        '''
        if not self.get_goc_enabled():
            self.set_goc_ein(goc=1)

        if self.minor == 3:
            logger.warn('ICE Firmware v0.3 reports wrong goc freq value.'\
                    ' Returning cached value.')
            try:
                return self.goc_freq
            except AttributeError:
                logger.warn('No cached value. Querying ICE. Value is junk')

        setting = self.goc_ein_get_freq_divisor()
        if self.minor == 1:
            NOMINAL = 2e6
        else:
            NOMINAL = 4e6
        freq_in_hz = NOMINAL / setting
        return freq_in_hz

    def _goc_freq_in_hz_to_divisor(self, freq_in_hz):
        if self.minor == 1:
            NOMINAL = 2e6
        else:
            NOMINAL = 4e6
        return NOMINAL / freq_in_hz;

    @min_proto_version("0.1")
    @capability('o')
    def goc_set_frequency(self, freq_in_hz):
        '''
        Sets the GOC frequency.
        '''
        if not self.get_goc_enabled():
            self.set_goc_ein(goc=1)

        # Send a 3-byte value N, where 2 MHz / N == clock speed
        self.goc_ein_set_freq_divisor(self._goc_freq_in_hz_to_divisor(freq_in_hz))

        self.goc_freq = freq_in_hz
        logger.debug("GOC frequency set to %f" % (freq_in_hz))

    @min_proto_version("0.2")
    @capability('O')
    def goc_get_onoff(self):
        '''
        Get the current ambient GOC power.
        '''
        if not self.get_goc_enabled():
            self.set_goc_ein(goc=1)

        self.min_version(0.2)
        resp = self.send_message_until_acked('O', struct.pack("B", ord('o')))
        if len(resp) != 1:
            raise self.FormatError("Wrong response length from `Oo': " + str(resp))
        onoff = struct.unpack("B", resp)[0]
        return bool(onoff)

    @min_proto_version("0.2")
    @capability('o')
    def goc_set_onoff(self, onoff):
        '''
        Turn the GOC light on or off.

        The GOC will blink as normal when goc_send is called, this simply sets
        the state of the GOC light when it's not doing anything else (e.g. so
        you can leave the light on for charging or something similar)
        '''
        if not self.get_goc_enabled():
            self.set_goc_ein(goc=1)

        self.min_version(0.2)
        msg = struct.pack("BB", ord('o'), onoff)
        self.send_message_until_acked('o', msg)

    ## I2C ##
    @min_proto_version("0.1")
    @capability('d')
    def i2c_send(self, addr, data):
        '''
        Sends an I2C message.

        Addr should be a single byte address.
        Data should be packed binary data, as returned by struct.pack

        The return value is the number of bytes actually sent *including the
        address byte*.

        Long messages may be fragmented between the ICE library and the ICE
        FPGA. On the I2C wire, this will appear as windows of time where the I2C
        clock is stretched for a period of time.  A faster baud rate between the
        PC host and the ICE FPGA will help mitigate this.
        '''

        msg = struct.pack("B", addr) + data
        return self._fragment_sender('d', msg)

    @min_proto_version("0.1")
    @capability('I')
    def i2c_get_speed(self):
        '''
        Get the clock speed of the ICE I2C driver in kHz.
        '''
        ack,msg = self.send_message('I', struct.pack("B", ord('c')))
        if ack == 0:
            if len(msg) != 1:
                raise self.FormatError
            return struct.unpack("B", msg)[0] * 2

        ret = ord(msg[0])
        msg = msg[1:]
        if ret == errno.ENODEV:
            # XXX Generalize me w.r.t. version?
            return 100
        else:
            raise self.ICE_Error("Unknown Error")

    @min_proto_version("0.1")
    @capability('i')
    def i2c_set_speed(self, speed):
        '''
        Set the clock speed of the ICE I2C driver in kHz.

        The accepted range of speeds is [2,400] kHz with steps of undefined
        increments. The actual set speed is returned.

        Raises an ICE_Error if the speed was not set.

        Note: This does *NOT* affect the clock speed of any M3 I2C drivers.
              That requires sending DMA messages to each of the M3 I2C
              controllers that you would like to change the speed of.
        '''
        if speed < 2:
            speed = 2
        elif speed > 400:
            speed = 400

        speed //= 2
        ack,msg = self.send_message('i', struct.pack("BB", ord('c'), speed))

        if ack == 0:
            return speed

        ret = ord(msg[0])
        msg = msg[1:]
        if ret == errno.EINVAL:
            raise self.ICE_Error("ICE reports: Invalid argument.")
        elif ret == errno.ENODEV:
            raise self.ICE_Error("Changing I2C speed not supported.")

    @min_proto_version("0.1")
    @capability('I')
    def i2c_get_address(self):
        '''
        Get the I2C address(es) of the ICE peripheral.
        '''
        resp = self.send_message_until_acked('I', struct.pack("B", ord('a')))
        if len(resp) != 2:
            raise self.FormatError("i2c address response should be 2 bytes")
        ones, zeros = struct.unpack("BB", resp)
        if ones == 0xff and zeros == 0xff:
            return None
        else:
            return self.masks_to_strings(ones, zeros, 8)

    @min_proto_version("0.1")
    @capability('i')
    def i2c_set_address(self, address=None):
        '''
        Set the I2C address(es) of the ICE peripheral.

        The ICE board will ACK messages sent to any address that matches the
        mask set by this function. The special character 'x' is used to signify
        don't-care bits. As example, to pretend to be the DSP layer:

           address = "1001 100x"

        Spaces are permitted and ignored. To disable this feature, set the
        address to None.

        Default Value: DISABLED.
        '''
        if address is None:
            ones, zeros = (0xff, 0xff)
        else:
            if len(address) != 8:
                raise self.FormatError("Address must be exactly 8 bits")
            ones, zeros = self.string_to_masks(address)
        self.send_message_until_acked('i', struct.pack("BBB", ord('a'), ones, zeros))

    ## MBus ##
    @min_proto_version("0.2")
    @capability('b')
    def mbus_send(self, addr, data):
        '''
        Sends an MBus message.

        Addr may be a short address or long address. In either case, it should
        be packed binary data (e.g. struct.pack or binascii.unhexlify('a5'))

        The return value is the number of bytes actually sent *including four
        bytes for the address, regardless of whether a short or long address was
        actually sent*.

        Long messages may be fragmented between the ICE library and the ICE
        FPGA. On the wire, this should not be noticeable as the PC<-->ICE bridge
        (3 MBaud) is much faster than the MBus. If this is an issue, you must
        keep the transaction size below the ICE fragmentation limit (less than
        255 bytes for combined address + data).
        '''
        self.min_version(0.2)
        if type(addr) != bytes:
            addr = bytes(addr, 'utf-8')
        if len(addr) > 4:
            raise self.FormatError("Address too long: " + str(addr) +\
                                    ' len:' + str(len(addr)))
        while len(addr) < 4:
            zero = bytes(1)
            addr = zero + addr
        msg = addr + data
        return self._fragment_sender('b', msg)

    @min_proto_version("0.3")
    @capability('m')
    def mbus_set_internal_reset(self, assert_reset):
        '''
        Control signal that holds ICE internal MBus in reset.

        While in reset, the COUT and DOUT signals are held high. This is useful
        for bootstrapping when multiple ICE boards are in a loop.
        '''
        self.min_version(0.3)
        self.send_message_until_acked('m', struct.pack("B"*(1+1),
            ord('r'),
            bool(assert_reset),
            ))

    @min_proto_version("0.2")
    @capability('m')
    def mbus_set_full_prefix(self, prefix=None):
        '''
        Set the full prefix(es) of the ICE peripheral.

        The ICE board will ACK messages sent to any address that matches the
        mask set by this function. The special character 'x' is used to signify
        don't-care bits.

        Spaces are permitted and ignored. To disable this feature, set the
        address to None.

        Default Value: DISABLED.
        '''
        self.min_version(0.2)
        if prefix is None:
            ones, zeros = (0xfffff, 0xfffff)
        else:
            if len(prefix) != 20:
                raise self.FormatError("Prefix must be exactly 20 bits")
            ones, zeros = self.string_to_masks(prefix)
        ones <<= 4
        zeros <<= 4
        self.send_message_until_acked('m', struct.pack("B"*(1+6),
            ord('l'),
            (ones >> 16) & 0xff,
            (ones >> 8) & 0xff,
            ones & 0xff,
            (zeros >> 16) & 0xff,
            (zeros >> 8) & 0xff,
            zeros & 0xff,
            ))

    @min_proto_version("0.2")
    @capability('M')
    def mbus_get_full_prefix(self):
        '''
        Get the full prefix(es) set for ICE.
        '''
        self.min_version(0.2)
        resp = self.send_message_until_acked('M', struct.pack("B", ord('l')))
        if len(resp) != 6:
            raise self.FormatError("Full prefix response should be 6 bytes")
        o_hig, o_mid, o_low, z_hig, z_mid, z_low = struct.unpack("BBBBBB", resp)
        ones = o_low | o_mid << 8 | o_hig << 16
        zeros = z_low | z_mid << 8 | z_hig << 16
        ones >>= 4
        zeros >>= 4
        if ones == 0xfffff and zeros == 0xfffff:
            return None
        else:
            return self.masks_to_strings(ones, zeros, 20)

    @min_proto_version("0.2")
    @capability('m')
    def mbus_set_short_prefix(self, prefix=None):
        '''
        Set the short prefix(es) of the ICE peripheral.

        Default Value: DISABLED.
        '''
        self.min_version(0.2)
        if prefix is None:
            ones, zeros = (0xf, 0xf)
        else:
            if prefix.startswith('0x'):
                try: # try converting from hex
                    prefix = "{0:b}".format( int(prefix,16) )
                except: 
                    raise self.FormatError("Malformed Prefix")
                logger.debug("Prefix parsed as: " + str(prefix))

            if len(prefix) != 4:
                raise self.FormatError("Prefix must be exactly 4 bits")

            ones, zeros = self.string_to_masks(prefix)

        self.send_message_until_acked('m', struct.pack("B"*(1+1),
            ord('s'),
            ones,
            ))

    @min_proto_version("0.2")
    @capability('M')
    def mbus_get_short_prefix(self):
        '''
        Get the short prefix(es) set for ICE.
        '''
        self.min_version(0.2)
        resp = self.send_message_until_acked('M', struct.pack("B", ord('s')))
        if len(resp) != 2:
            raise self.FormatError("Full prefix response should be 2 bytes")
        ones, zeros = struct.unpack("BB", resp)
        ones >>= 4
        zeros >>= 4
        if ones == 0xf and zeros == 0xf:
            return None
        else:
            return self.masks_to_strings(ones, zeros, 4)

    @min_proto_version("0.3")
    @capability('m')
    def mbus_set_snoop(self, enable, filter_prefix=None):
        '''
        Enable snooping of all traffic. The optional filter runs in software to limit reported messages.

        Default Value: DISABLED.
        '''
        self.min_version(0.3)
        enable = bool(enable)
        if filter_prefix is not None:
            raise NotImplementedError
        self.send_message_until_acked('m', struct.pack("B"*(1+1),
            ord('S'),
            enable,
            ))

    @min_proto_version("0.3")
    @capability('M')
    def mbus_get_snoop(self, return_filter=False):
        '''
        Return whether snooping is enabled.
        '''
        self.min_version(0.3)
        resp = self.send_message_until_acked('M', struct.pack("B", ord('S')))
        if len(resp) != 1:
            raise self.FormatError("Snoop enabled response should be 1 byte")
        enabled = bool(struct.unpack("B", resp))

        if return_filter:
            raise NotImplementedError
        return enabled

    @min_proto_version("0.2")
    @capability('m')
    def mbus_set_broadcast_channel_mask(self, mask=None):
        '''
        Set the broadcast mask for ICE board.

        The ICE board will report and ACK any messages sent to broadcast
        channels that match the mask set by this function. The special character 'x' is
        used to signify don't-care bits.

        Spaces are permitted and ignored. To disable this feature, set the
        address to None.

        Default Value: DISABLED.
        '''
        self.min_version(0.2)
        if mask is None:
            ones, zeros = (0xf, 0xf)
        else:
            if len(mask) != 4:
                raise self.FormatError("Prefix must be exactly 4 bits")
            ones, zeros = self.string_to_masks(mask)
        self.send_message_until_acked('m', struct.pack("B"*(1+2),
            ord('b'),
            ones,
            zeros,
            ))

    @min_proto_version("0.2")
    @capability('M')
    def mbus_get_broadcast_channel_mask(self):
        '''
        Get the broadcast mask for ICE.
        '''
        self.min_version(0.2)
        resp = self.send_message_until_acked('M', struct.pack("B", ord('b')))
        if len(resp) != 2:
            raise self.FormatError("Broadcast mask response should be 2 bytes")
        ones, zeros = struct.unpack("BB", resp)
        if ones == 0xf and zeros == 0xf:
            return None
        else:
            return self.masks_to_strings(ones, zeros, 4)

    @min_proto_version("0.2")
    @capability('m')
    def mbus_set_broadcast_channel_snoop_mask(self, mask=None):
        '''
        Set the broadcast snoop mask for ICE board.

        The ICE board will report, but not ACK, any messages sent to broadcast
        channels that match the mask set by this function. The special character 'x' is
        used to signify don't-care bits.

        Spaces are permitted and ignored. To disable this feature, set the
        address to None.

        Default Value: DISABLED.
        '''
        self.min_version(0.2)
        if mask is None:
            ones, zeros = (0xf, 0xf)
        else:
            if len(mask) != 4:
                raise self.FormatError("Prefix must be exactly 4 bits")
            ones, zeros = self.string_to_masks(mask)
        self.send_message_until_acked('m', struct.pack("B"*(1+2),
            ord('B'),
            ones,
            zeros,
            ))

    @min_proto_version("0.2")
    @capability('M')
    def mbus_get_broadcast_channel_snoop_mask(self):
        '''
        Get the broadcast snoop mask for ICE.
        '''
        self.min_version(0.2)
        resp = self.send_message_until_acked('M', struct.pack("B", ord('B')))
        if len(resp) != 2:
            raise self.FormatError("Broadcast mask response should be 2 bytes")
        ones, zeros = struct.unpack("BB", resp)
        if ones == 0xf and zeros == 0xf:
            return None
        else:
            return self.masks_to_strings(ones, zeros, 4)

    @min_proto_version("0.2")
    @capability('M')
    def mbus_get_master_onoff(self):
        '''
        Get whether ICE is acting as MBus master node.
        '''
        self.min_version(0.2)
        resp = self.send_message_until_acked('M', struct.pack("B", ord('m')))
        if len(resp) != 1:
            raise self.FormatError("Wrong response length from `Mm': " + str(resp))
        onoff = struct.unpack("B", resp)[0]
        return bool(onoff)

    @min_proto_version("0.2")
    @capability('m')
    def mbus_set_master_onoff(self, onoff):
        '''
        Set whether ICE acts as MBus master node.

        DEFAULT: OFF
        '''
        self.min_version(0.2)
        if isinstance(onoff, str):
            onoff = onoff.lower()
            onoff = True if onoff in ['on'] else False
        elif isinstance(onoff, bool):
            pass
        else: raise Exception("Bad arg for " + __name__ )

        msg = struct.pack("BB", ord('m'), onoff)
        self.send_message_until_acked('m', msg)

    @min_proto_version("0.2")
    @capability('M')
    def mbus_get_clock(self):
        '''
        Get ICE MBus clock speed. Only meaningful if ICE is MBus master.
        '''
        self.min_version(0.2)
        raise NotImplementedError
        #resp = self.send_message_until_acked('M', struct.pack("B", ord('c')))
        #if len(resp) != 1:
        #    raise self.FormatError("Wrong response length from `Mc': " + str(resp))
        #onoff = struct.unpack("B", resp)[0]
        #return bool(onoff)
        #return resp

    @min_proto_version("0.2")
    @capability('m')
    def mbus_set_clock(self, clock_speed):
        '''
        Set ICE MBus clock speed. Only meaningful if ICE is MBus master.

        DEFAULT: XXX
        '''
        self.min_version(0.2)
        #msg = struct.pack("BB", ord('c'), onoff)
        #self.send_message_until_acked('m', msg)
        raise NotImplementedError

    @min_proto_version("0.2")
    @capability('M')
    def mbus_get_should_interrupt(self):
        '''
        Get ICE MBus should interrupt setting.

        TODO: Fix interface (enums?)
        '''
        self.min_version(0.2)
        resp = self.send_message_until_acked('M', struct.pack("B", ord('i')))
        resp = ord(resp)
        #if len(resp) != 1:
        #    raise self.FormatError("Wrong response length from `Mc': " + str(resp))
        #onoff = struct.unpack("B", resp)[0]
        #return bool(onoff)
        return resp

    @min_proto_version("0.2")
    @capability('m')
    def mbus_set_should_interrupt(self, should_interrupt):
        '''
        Set ICE MBus should interrupt setting.

        DEFAULT: Off
        '''
        self.min_version(0.2)
        msg = struct.pack("BB", ord('i'), should_interrupt)
        self.send_message_until_acked('m', msg)

    @min_proto_version("0.2")
    @capability('M')
    def mbus_get_use_priority(self):
        '''
        Get ICE MBus use priority setting.

        TODO: Fix interface (enums?)
        '''
        self.min_version(0.2)
        resp = self.send_message_until_acked('M', struct.pack("B", ord('p')))
        resp = ord(resp)
        #if len(resp) != 1:
        #    raise self.FormatError("Wrong response length from `Mc': " + str(resp))
        #onoff = struct.unpack("B", resp)[0]
        #return bool(onoff)
        return resp

    @min_proto_version("0.2")
    @capability('m')
    def mbus_set_use_priority(self, use_priority):
        '''
        Set ICE MBus use priority setting.

        DEFAULT: Off
        '''
        self.min_version(0.2)
        msg = struct.pack("BB", ord('p'), use_priority)
        self.send_message_until_acked('m', msg)

    ## EIN DEBUG ##
    EIN_DEFAULT_DIVISOR = 0xFA0

    @min_proto_version("0.2")
    @capability('f')
    def ein_send(self, msg):
        '''
        Sends a message via the EIN Debug port.

        Takes a raw byte stream (e.g. binascii.unhexlify("aa")).
        Returns the number of bytes actually sent.

        Long messages may be fragmented between the ICE library and the ICE
        FPGA. These fragments will be combined on the ICE board. There should be
        no interruption in message transmission.
        '''
        if not self.get_ein_enabled():
            self.set_goc_ein(ein=1)

        self.min_version(0.2)
        ret = self._fragment_sender('f', msg)
        return ret

    ## GPIO ##
    # XXX TODO XXX: parameter based method version selection
    GPIO_INPUT = 0
    GPIO_OUTPUT = 1
    GPIO_TRISTATE = 2

    def gpio_get_level(self, gpio_idx):
        '''
        Query whether a gpio is high or low. (high=True)
        '''
        if self.minor == 1:
            return self.gpio_get_level_0_1(gpio_idx)
        else:
            return self.gpio_get_level_0_2(gpio_idx)

    def gpio_get_direction(self, gpio_idx):
        '''
        Query gpio pin setup.

        Returns one of:
            ICE.GPIO_INPUT
            ICE.GPIO_OUTPUT
            ICE.GPIO_TRISTATE
        '''
        if self.minor == 1:
            return self.gpio_get_direction_0_1(gpio_idx)
        else:
            return self.gpio_get_direction_0_2(gpio_idx)

    def gpio_set_level(self, gpio_idx, level):
        '''
        Set gpio level. (high=True)
        '''
        if self.minor == 1:
            return self.gpio_set_level_0_1(gpio_idx, level)
        else:
            return self.gpio_set_level_0_2(gpio_idx, level)

    def gpio_set_direction(self, gpio_idx, direction):
        '''
        Setup a GPIO pin.
        '''
        if direction not in (ICE.GPIO_INPUT, ICE.GPIO_OUTPUT, ICE.GPIO_TRISTATE):
            raise self.ParameterError("Unknown direction: " + str(direction))
        if self.minor == 1:
            return self.gpio_set_direction_0_1(gpio_idx, direction)
        else:
            return self.gpio_set_direction_0_2(gpio_idx, direction)

    @min_proto_version("0.1")
    @max_proto_version("0.1")
    @capability('G')
    def gpio_get_level_0_1(self, gpio_idx):
        resp = self.send_message_until_acked('G',
                struct.pack('BB', ord('l'), gpio_idx))
        if len(resp) != 1:
            raise self.FormatError("Too long of a response from `Gl#':" + str(resp))

        return bool(struct.unpack("B", resp)[0])

    @min_proto_version("0.1")
    @max_proto_version("0.1")
    @capability('G')
    def gpio_get_direction_0_1(self, gpio_idx):
        resp = self.send_message_until_acked('G',
                struct.pack('BB', ord('d'), gpio_idx))
        if len(resp) != 1:
            raise self.FormatError("Too long of a response from `Gd#':" + str(resp))

        direction = struct.unpack("B", resp)[0]
        if direction not in (ICE.GPIO_INPUT, ICE.GPIO_OUTPUT, ICE.GPIO_TRISTATE):
            raise self.FormatError("Unknown direction: " + str(direction))

        return direction

    @min_proto_version("0.1")
    @max_proto_version("0.1")
    @capability('g')
    def gpio_set_level_0_1(self, gpio_idx, level):
        self.send_message_until_acked('g',
                struct.pack('BBB', ord('l'), gpio_idx, level))

    @min_proto_version("0.1")
    @max_proto_version("0.1")
    @capability('g')
    def gpio_set_direction_0_1(self, gpio_idx, direction):
        self.send_message_until_acked('g',
                struct.pack('BBB', ord('d'), gpio_idx, direction))

    def _gpio_get_level_0_2(self):
        resp = self.send_message_until_acked('G', struct.pack('B', ord('l')))
        if len(resp) != 3:
            raise self.FormatError("Bad response from `Gl':" + str(resp))
        high,mid,low = map(ord, resp)
        return low | (mid << 8) | (high << 16)

    @min_proto_version("0.2")
    @capability('G')
    def gpio_get_level_0_2(self, gpio_idx):
        if gpio_idx >= 24:
            raise self.ParameterError("Request for illegal gpio idx")
        return (self._gpio_get_level_0_2() >> gpio_idx) & 0x1

    def _gpio_get_direction_0_2(self):
        resp = self.send_message_until_acked('G', struct.pack('B', ord('d')))
        if len(resp) != 3:
            raise self.FormatError("Bad response from `Gd#':" + str(resp))
        high,mid,low = map(ord, resp)
        return low | (mid << 8) | (high << 16)

    @min_proto_version("0.2")
    @capability('G')
    def gpio_get_direction_0_2(self, gpio_idx):
        if gpio_idx >= 24:
            raise self.ParameterError("Request for illegal gpio idx")

        if ((self._gpio_get_direction_0_2() >> gpio_idx) & 0x1) == 0:
            return ICE.GPIO_INPUT
        else:
            return ICE.GPIO_OUTPUT

    @min_proto_version("0.2")
    @capability('g')
    def gpio_set_level_0_2(self, gpio_idx, level):
        mask = self._gpio_get_level_0_2()
        if level:
            mask |= (1 << gpio_idx)
        else:
            mask &= ~(1 << gpio_idx)
        self.send_message_until_acked('g',
                struct.pack('BBBB', ord('l'),
                    (mask >> 16) & 0xff,
                    (mask >> 8) & 0xff,
                    mask & 0xff))

    @min_proto_version("0.2")
    @capability('g')
    def gpio_set_direction_0_2(self, gpio_idx, direction):
        mask = self._gpio_get_direction_0_2()
        if direction == ICE.GPIO_OUTPUT:
            mask |= (1 << gpio_idx)
        elif direction in (ICE.GPIO_INPUT, ICE.GPIO_TRISTATE):
            mask &= ~(1 << gpio_idx)
        else:
            raise self.ParameterError("Illegal GPIO direction")
        self.send_message_until_acked('g',
                struct.pack('BBBB', ord('d'),
                    (mask >> 16) & 0xff,
                    (mask >> 8) & 0xff,
                    mask & 0xff))

    @min_proto_version("0.2")
    @capability('G')
    def gpio_get_interrupt_enable_mask(self):
        resp = self.send_message_until_acked('G', struct.pack('B', ord('i')))
        if len(resp) != 3:
            raise self.FormatError("Bad response from `Gi':" + str(resp))
        high,mid,low = map(ord, resp)
        return low | (mid << 8) | (high << 16)

    @min_proto_version("0.2")
    @capability('g')
    def gpio_set_interrupt_enable_mask(self, mask):
        self.send_message_until_acked('g',
                struct.pack('BBBB', ord('i'),
                    (mask >> 16) & 0xff,
                    (mask >> 8) & 0xff,
                    mask & 0xff))

    ## POWER ##
    POWER_0P6 = 0
    POWER_1P2 = 1
    POWER_VBATT = 2
    POWER_GOC = 3

    POWER_0P6_DEFAULT = 0.675
    POWER_1P2_DEFAULT = 1.2
    POWER_VBATT_DEFAULT = 3.8

    @min_proto_version("0.1")
    @capability('P')
    def power_get_voltage(self, rail):
        '''
        Query the current voltage setting of a power rail.

        The `rail' argument must be one of:
            ICE.POWER_0P6
            ICE.POWER_1P2
            ICE.POWER_VBATT
        '''
        if rail not in (ICE.POWER_0P6, ICE.POWER_1P2, ICE.POWER_VBATT):
            raise self.ParameterError("Invalid rail: " + str(rail))

        logger.warn("ICE Firmware <= 0.3 cannot query voltage. Returning cached value.")
        try:
            raw = getattr(self, 'power_{}'.format(rail))
        except AttributeError:
            logger.warn("No cached value, returning default")
            raw = (1 - 0.537) / 0.0185

        #resp = self.send_message_until_acked('P', struct.pack("BB", ord('v'), rail))
        #if len(resp) != 2:
        #    raise self.FormatError("Wrong response length from `Pv#':" + str(resp))
        #rail, raw = struct.unpack("BB", resp)

        # Vout = (0.537 + 0.0185 * v_set) * Vdefault
        default_voltage = (ICE.POWER_0P6_DEFAULT, ICE.POWER_1P2_DEFAULT,
                ICE.POWER_VBATT_DEFAULT)[rail]
        vout = (0.537 + 0.0185 * raw) * default_voltage
        return vout

    # didn't actually work until v0.5
    @min_proto_version("0.5")
    @capability('P')
    def power_get_onoff(self, rail):
        '''
        Query the current on/off setting of a power rail.

        Returns a boolean, on=True.
        '''
        if rail not in (ICE.POWER_0P6, ICE.POWER_1P2, ICE.POWER_VBATT, ICE.POWER_GOC):
            raise self.ParameterError("Invalid rail: " + str(rail))

        resp = self.send_message_until_acked('P', struct.pack("BB", ord('o'), rail))
        if len(resp) != 1:
            raise self.FormatError("Too long of a response from `Po#':" + str(resp))
        onoff = struct.unpack("B", resp)[0]
        return bool(onoff)

    @min_proto_version("0.1")
    @capability('p')
    def power_set_voltage(self, rail, output_voltage):
        '''
        Set the voltage setting of a power rail. Units are V.
        '''
        if rail not in (ICE.POWER_0P6, ICE.POWER_1P2, ICE.POWER_VBATT):
            raise self.ParameterError("Invalid rail: " + str(rail))

        # Vout = (0.537 + 0.0185 * v_set) * Vdefault
        output_voltage = float(output_voltage)
        default_voltage = (ICE.POWER_0P6_DEFAULT, ICE.POWER_1P2_DEFAULT,
                ICE.POWER_VBATT_DEFAULT)[rail]
        vset = ((output_voltage / default_voltage) - 0.537) / 0.0185
        vset = int(vset)
        if (vset < 0) or (vset > 255):
            raise self.ParameterError("Voltage exceeds range. vset: " + str(vset))

        self.send_message_until_acked('p', struct.pack("BBB", ord('v'), rail, vset))
        setattr(self, 'power_{}'.format(rail), vset)

    @min_proto_version("0.1")
    @capability('p')
    def power_set_onoff(self, rail, onoff):
        '''
        Turn a power rail on or off (on=True).
        '''
        if rail not in (ICE.POWER_0P6, ICE.POWER_1P2, ICE.POWER_VBATT, ICE.POWER_GOC):
            raise self.ParameterError("Invalid rail: " + str(rail))

        self.send_message_until_acked('p', struct.pack("BBB", ord('o'), rail, onoff))

if __name__ == '__main__':
    logger.setLevel(level=logging.DEBUG)

